

import random
from altgraph.Graph import Graph
import copy
import itertools

from Class.exceptions import ValidationException
from Class.itemClass import KH2Item
from Class.newLocationClass import KH2Location
from List.NewLocationList import get_all_parent_edge_requirements, Locations
from List.configDict import ItemAccessibilityOption, locationType
from List.inventory import storyunlock, keyblade, proof, form, magic, misc
from List.location import simulatedtwilighttown as stt
from List.location import worldthatneverwas as twtnw
from List.location.graph import RequirementFunction
from Module.RandomizerSettings import RandomizerSettings
from Module.itemPlacementRestriction import ItemPlacementHelpers
from Module.newRandomize import Randomizer, SynthesisRecipe


class ValidationResult:

    def __init__(self):
        self.any_percent = False
        self.full_clear = False


class LocationInformedSeedValidator:

    def __init__(self):
        self.location_requirements: dict[KH2Location, list[RequirementFunction]] = {}
        self.human_readable_lock_list: dict[locationType,Graph] = {}
        self.location_spheres: dict[KH2Location, int] = {}

    def populate_possible_locking_item_list(self, regular_rando: bool, keyblade_locking: bool):
        # STT
        stt_graph = Graph()
        stt_graph.add_node("STT",[storyunlock.NaminesSketches])
        stt_graph.add_node("STTChests",[keyblade.BondOfFlame])
        stt_graph.add_edge("STT","STTChests")
        self.human_readable_lock_list[locationType.STT] = stt_graph
        # HB
        hb_graph = Graph()
        hb_graph.add_node("HB1",[storyunlock.MembershipCard])
        hb_graph.add_node("HB2",[storyunlock.MembershipCard,storyunlock.MembershipCard])
        hb_graph.add_edge("HB1","HB2")
        if keyblade_locking:
            hb_graph.add_node("HBChests",[keyblade.SleepingLion])
            hb_graph.add_node("CoR",[keyblade.WinnersProof])
            hb_graph.add_edge("HBChests","HB2")
            hb_graph.add_edge("HBChests","CoR")
            hb_graph.add_edge("CoR","HBChests")
            hb_graph.add_edge("HB1","HBChests")
            hb_graph.add_edge("HB2","HBChests")
            hb_graph.add_edge("HB2","CoR")
        if regular_rando:
            hb_graph.add_node("PoP",[proof.ProofOfPeace])
            hb_graph.add_edge("HB2","PoP")
            if keyblade_locking:
                hb_graph.add_edge("PoP","HBChests")
                hb_graph.add_edge("CoR","PoP")
                hb_graph.add_edge("PoP","CoR")
        self.human_readable_lock_list[locationType.HB] = hb_graph
        # OC
        oc_graph = Graph()
        oc_graph.add_node("OC1",[storyunlock.BattlefieldsOfWar])
        oc_graph.add_node("OC2",[storyunlock.BattlefieldsOfWar,storyunlock.BattlefieldsOfWar])
        oc_graph.add_edge("OC1","OC2")
        if keyblade_locking:
            oc_graph.add_node("OCChests",[keyblade.HerosCrest])
            oc_graph.add_edge("OC1","OCChests")
            oc_graph.add_edge("OC2","OCChests")
            oc_graph.add_edge("OCChests","OC2")
        self.human_readable_lock_list[locationType.OC] = oc_graph
        # LoD
        lod_graph = Graph()
        lod_graph.add_node("LoD1",[storyunlock.SwordOfTheAncestor])
        lod_graph.add_node("LoD2",[storyunlock.SwordOfTheAncestor,storyunlock.SwordOfTheAncestor])
        lod_graph.add_edge("LoD1","LoD2")
        if keyblade_locking:
            lod_graph.add_node("LoDChests",[keyblade.HiddenDragon])
            lod_graph.add_edge("LoD1","LoDChests")
            lod_graph.add_edge("LoD2","LoDChests")
            lod_graph.add_edge("LoDChests","LoD2")
        self.human_readable_lock_list[locationType.LoD] = lod_graph
        # PL
        pl_graph = Graph()
        pl_graph.add_node("PL1",[storyunlock.ProudFang])
        pl_graph.add_node("PL2",[storyunlock.ProudFang,storyunlock.ProudFang])
        pl_graph.add_edge("PL1","PL2")
        if keyblade_locking:
            pl_graph.add_node("PLChests",[keyblade.CircleOfLife])
            pl_graph.add_edge("PL1","PLChests")
            pl_graph.add_edge("PL2","PLChests")
            pl_graph.add_edge("PLChests","PL2")
        self.human_readable_lock_list[locationType.PL] = pl_graph
        # HT
        ht_graph = Graph()
        ht_graph.add_node("HT1",[storyunlock.BoneFist])
        ht_graph.add_node("HT2",[storyunlock.BoneFist,storyunlock.BoneFist])
        ht_graph.add_edge("HT1","HT2")
        if keyblade_locking:
            ht_graph.add_node("HTChests",[keyblade.DecisivePumpkin])
            ht_graph.add_edge("HT1","HTChests")
            ht_graph.add_edge("HT2","HTChests")
            ht_graph.add_edge("HTChests","HT2")
        self.human_readable_lock_list[locationType.HT] = ht_graph
        # SP
        sp_graph = Graph()
        sp_graph.add_node("SP1",[storyunlock.IdentityDisk])
        sp_graph.add_node("SP2",[storyunlock.IdentityDisk,storyunlock.IdentityDisk])
        sp_graph.add_edge("SP1","SP2")
        if keyblade_locking:
            sp_graph.add_node("SPChests",[keyblade.PhotonDebugger])
            sp_graph.add_edge("SP1","SPChests")
            sp_graph.add_edge("SP2","SPChests")
            sp_graph.add_edge("SPChests","SP2")
        self.human_readable_lock_list[locationType.SP] = sp_graph
        # Drives
        drive_graph = Graph()
        drive_graph.add_node("Valor",[form.ValorForm])
        drive_graph.add_node("Wisdom",[form.WisdomForm])
        drive_graph.add_node("Limit",[form.LimitForm])
        drive_graph.add_node("Master",[form.MasterForm])
        drive_graph.add_node("Final",[form.FinalForm])
        for n in drive_graph.node_list():
            for m in drive_graph.node_list():
                if n is not m:
                    drive_graph.add_edge(n,m)
        self.human_readable_lock_list[locationType.FormLevel] = drive_graph
        # TT
        tt_graph = Graph()
        tt_graph.add_node("TT1",[storyunlock.IceCream])
        tt_graph.add_node("TT2",[storyunlock.IceCream,storyunlock.IceCream])
        tt_graph.add_node("TT3",[storyunlock.IceCream,storyunlock.IceCream,storyunlock.IceCream])
        tt_graph.add_edge("TT1","TT2")
        tt_graph.add_edge("TT2","TT3")
        if keyblade_locking:
            tt_graph.add_node("TTChests",[keyblade.Oathkeeper])
            tt_graph.add_edge("TT1","TTChests")
            tt_graph.add_edge("TT2","TTChests")
            tt_graph.add_edge("TT3","TTChests")
            tt_graph.add_edge("TTChests","TT2")
            tt_graph.add_edge("TTChests","TT3")
        self.human_readable_lock_list[locationType.TT] = tt_graph
        # BC
        bc_graph = Graph()
        bc_graph.add_node("BC1",[storyunlock.BeastsClaw])
        bc_graph.add_node("BC2",[storyunlock.BeastsClaw,storyunlock.BeastsClaw])
        bc_graph.add_edge("BC1","BC2")
        if keyblade_locking:
            bc_graph.add_node("BCChests",[keyblade.RumblingRose])
            bc_graph.add_edge("BC1","BCChests")
            bc_graph.add_edge("BC2","BCChests")
            bc_graph.add_edge("BCChests","BC2")
        self.human_readable_lock_list[locationType.BC] = bc_graph
        # HAW
        haw_graph = Graph()
        haw_graph.add_node("Page1",[misc.TornPages])
        haw_graph.add_node("Page2",[misc.TornPages,misc.TornPages])
        haw_graph.add_node("Page3",[misc.TornPages,misc.TornPages,misc.TornPages])
        haw_graph.add_node("Page4",[misc.TornPages,misc.TornPages,misc.TornPages,misc.TornPages])
        haw_graph.add_node("Page5",[misc.TornPages,misc.TornPages,misc.TornPages,misc.TornPages,misc.TornPages])
        if keyblade_locking:
            haw_graph.add_node("PoohChests",[keyblade.SweetMemories])
            haw_graph.add_edge("PoohChests","Page1")
        haw_graph.add_edge("Page1","Page2")
        haw_graph.add_edge("Page2","Page3")
        haw_graph.add_edge("Page3","Page4")
        haw_graph.add_edge("Page4","Page5")
        self.human_readable_lock_list[locationType.HUNDREDAW] = haw_graph
        # DC
        dc_graph = Graph()
        dc_graph.add_node("DC1", [storyunlock.RoyalSummons])
        dc_graph.add_node("DC2", [storyunlock.RoyalSummons, storyunlock.RoyalSummons])
        dc_graph.add_node("PoC",[proof.ProofOfConnection])
        dc_graph.add_edge("DC1","DC2")
        dc_graph.add_edge("DC2","PoC")
        if keyblade_locking:
            dc_graph.add_node("DCChests",[keyblade.Monochrome])
            dc_graph.add_edge("DC1","DCChests")
            dc_graph.add_edge("DC2","DCChests")
            dc_graph.add_edge("DCChests","DC2")
            dc_graph.add_edge("DCChests","PoC")
        self.human_readable_lock_list[locationType.DC] = dc_graph
        # PR
        pr_graph = Graph()
        pr_graph.add_node("PR1",[storyunlock.SkillAndCrossbones])
        pr_graph.add_node("PR2",[storyunlock.SkillAndCrossbones,storyunlock.SkillAndCrossbones])
        pr_graph.add_edge("PR1","PR2")
        if keyblade_locking:
            pr_graph.add_node("PRChests",[keyblade.FollowTheWind])
            pr_graph.add_edge("PR1","PRChests")
            pr_graph.add_edge("PR2","PRChests")
            pr_graph.add_edge("PRChests","PR2")
        self.human_readable_lock_list[locationType.PR] = pr_graph
        # TWTNW
        twtnw_graph = Graph()
        twtnw_graph.add_node("TWTNW1",[storyunlock.WayToTheDawn])
        twtnw_graph.add_node("TWTNW2",[storyunlock.WayToTheDawn,storyunlock.WayToTheDawn])
        twtnw_graph.add_edge("TWTNW1","TWTNW2")
        if keyblade_locking:
            twtnw_graph.add_node("TWTNWChests",[keyblade.TwoBecomeOne])
            twtnw_graph.add_edge("TWTNW1","TWTNWChests")
            twtnw_graph.add_edge("TWTNW2","TWTNWChests")
            twtnw_graph.add_edge("TWTNWChests","TWTNW2")
        self.human_readable_lock_list[locationType.TWTNW] = twtnw_graph
        # Atlantica
        atlantica_graph = Graph()
        atlantica_graph.add_node("Song4",[magic.Magnet,magic.Magnet])
        atlantica_graph.add_node("Song5",[magic.Magnet,magic.Magnet,magic.Thunder,magic.Thunder,magic.Thunder])
        atlantica_graph.add_edge("Song4","Song5")
        if not regular_rando:
            atlantica_graph.add_node("Song2",[magic.Magnet])
            atlantica_graph.add_edge("Song2","Song4")
        self.human_readable_lock_list[locationType.Atlantica] = atlantica_graph
        # AG
        if regular_rando:
            ag_graph = Graph()
            ag_graph.add_node("AG1",[storyunlock.Scimitar])
            ag_graph.add_node("AG2",[storyunlock.Scimitar,storyunlock.Scimitar,magic.Fire,magic.Blizzard,magic.Thunder])
            ag_graph.add_edge("AG1","AG2")
            if keyblade_locking:
                ag_graph.add_node("AGChests",[keyblade.WishingLamp])
                ag_graph.add_edge("AG1","AGChests")
                ag_graph.add_edge("AG2","AGChests")
                ag_graph.add_edge("AGChests","AG2")
            self.human_readable_lock_list[locationType.Agrabah] = ag_graph
        else:
            ag_graph = Graph()
            ag_graph.add_node("AG1",[storyunlock.Scimitar]) # there are no first visit checks in reverse without the key
            ag_graph.add_node("AG1.5",[magic.Fire,magic.Blizzard,magic.Thunder])
            ag_graph.add_node("AG2",[storyunlock.Scimitar,storyunlock.Scimitar])
            ag_graph.add_edge("AG1","AG1.5")
            ag_graph.add_edge("AG1.5","AG2")
            if keyblade_locking:
                ag_graph.add_node("AGChests",[storyunlock.Scimitar,keyblade.WishingLamp])
                ag_graph.add_edge("AG1","AGChests")
                ag_graph.add_edge("AG1.5","AGChests")
                ag_graph.add_edge("AG2","AGChests")
                ag_graph.add_edge("AGChests","AG1.5")
                ag_graph.add_edge("AGChests","AG2")
            self.human_readable_lock_list[locationType.Agrabah] = ag_graph

    def generate_locking_item_ids(self):
        def random_walk(graph: Graph):
            current_node_list = []
            current_item_data = []
            current_node = next((n for n in graph.node_list() if graph.inc_degree(n)==0),None)
            if current_node is None:
                # if there is no obvious start node, pick one at random
                current_node = random.choice(graph.node_list())

            while True:
                # populate the data for the current node
                current_node_list.append(current_node)
                current_item_data.append(graph.node_data(current_node))
                # find out edge nodes that we haven't traversed yet
                candidate_nodes = [n for n in graph.out_nbrs(current_node) if n not in current_node_list]
                if len(candidate_nodes)==0:
                    break
                current_node = random.choice(candidate_nodes)
                # repeat
            if len(graph.node_list()) != len(current_item_data):
                return None
            return current_item_data


        item_id_lock_list: dict[locationType,list[list[int]]] = {}
        for location_type,lock_graph in self.human_readable_lock_list.items():
            # convert this from graph to randomly ordered item id list
            random_ordered_requirements = None
            while random_ordered_requirements is None:
                random_ordered_requirements = random_walk(lock_graph)

            new_list = []
            for lock_item_list in random_ordered_requirements:
                new_list.append([])
                for lock_item in lock_item_list:
                    new_list[-1].append(lock_item.id)
            item_id_lock_list[location_type] = new_list
        return item_id_lock_list

    @staticmethod
    def evaluate(inventory: list[int], reqs_list: list[RequirementFunction]) -> bool:
        return all([r(inventory) for r in reqs_list])

    def is_location_available(self, inventory: list[int], location: KH2Location) -> bool:
        return self.evaluate(inventory, self.location_requirements[location])

    def prepare_requirements_list(self, location_lists: list[Locations], synthesis_recipes: list[SynthesisRecipe]):
        self.location_requirements.clear()
        for locations in location_lists:
            for node_id in locations.node_ids():
                parent_requirement_function = get_all_parent_edge_requirements(node_id, locations.location_graph)
                for location in locations.locations_for_node(node_id):
                    if location not in self.location_requirements:
                        self.location_requirements[location] = []
                    self.location_requirements[location].append(parent_requirement_function)

        for loc, requirements in self.location_requirements.items():
            if locationType.SYNTH in loc.LocationTypes:
                # this is a synth location, we need to get its recipe to know what locks it logically
                recipe = next((r for r in synthesis_recipes if r.location == loc), None)
                # if we don't have recipes yet, we can't validate that yet
                if recipe:
                    for recipe_requirement in recipe.requirements:
                        synth_item = recipe_requirement.synth_item
                        # now that we know what synth item is in the recipe, we can determine what to add to the logic
                        requirements.append(ItemPlacementHelpers.make_synth_requirement(synth_item))

    def prep_requirements_list(self, settings: RandomizerSettings, randomizer: Randomizer):
        location_lists = []
        if settings.regular_rando:
            location_lists.append(randomizer.regular_locations)
        if settings.reverse_rando:
            location_lists.append(randomizer.reverse_locations)
        self.prepare_requirements_list(location_lists, randomizer.synthesis_recipes)
        self.populate_possible_locking_item_list(settings.regular_rando, settings.keyblades_unlock_chests)

    def validate_seed(
            self,
            settings: RandomizerSettings,
            randomizer: Randomizer,
            verbose: bool = True
    ) -> list[KH2Location]:
        self.prep_requirements_list(settings, randomizer)

        location_requirements = copy.deepcopy(self.location_requirements)

        results = ValidationResult()
        inventory = [item_id for item_id in randomizer.starting_item_ids]
        if len(randomizer.shop_items) > 0:
            for shop_item in randomizer.shop_items:
                inventory.append(shop_item.Id)

        changed = True
        depth = 0
        while changed:
            depth += 1
            final_xem_in_list = False
            if not results.any_percent:
                for location, requirements in location_requirements.items():
                    if location.name() == twtnw.CheckLocation.FinalXemnas:
                        final_xem_in_list = True
                        break
                if not final_xem_in_list:
                    results.any_percent = True

            if len(location_requirements) == 0:
                if verbose:
                    print(f"Logic Depth {depth}")
                results.full_clear = True
                break

            changed = False
            locations_to_remove = []
            items_to_add_to_inventory = []
            for location, requirements in location_requirements.items():
                if self.evaluate(inventory, requirements):
                    # find assigned item to location
                    for assignment in randomizer.assignments:
                        # if assignment is one of the struggle win/lose items, only count one, and not count the second.
                        if assignment.location.name() == stt.CheckLocation.StruggleWinnerChampionBelt:
                            continue
                        if location == assignment.location:
                            items_to_add_to_inventory.append(assignment.item.Id)
                            if assignment.item2 is not None:
                                items_to_add_to_inventory.append(assignment.item2.Id)
                            break
                    locations_to_remove.append(location)
                    self.location_spheres[location] = depth-1
                    changed = True
            for shop_item in locations_to_remove:
                location_requirements.pop(shop_item)
            inventory.extend(items_to_add_to_inventory)

        if (settings.item_accessibility == ItemAccessibilityOption.ALL and results.full_clear) \
                or (settings.item_accessibility == ItemAccessibilityOption.BEATABLE and results.any_percent):
            # we all good
            if verbose:
                print(f"Unreachable locations {len(location_requirements)}")
            return self.location_spheres
        else:
            if verbose:
                print("Failed seed, trying again")
            
            # print(inventory)
            # for loc_req in location_requirements.items():
            #     print(loc_req[0].Description)
            #     for assignment in randomizer.assignments:
            #         if loc_req[0] == assignment.location:
            #             print(f"---{assignment.item.Name}")
            raise ValidationException(f"Completion checking failed to collect {len(location_requirements)} items")



from enum import Enum

from List.configDict import locationType, itemType
from List.inventory import misc, magic, keyblade, ability, form
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, item_bonus, stat_bonus, hybrid_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    DisneyCastleCourtyard = "DC Courtyard"
    DisneyCastleCourtyardChests = "DC Courtyard Chests"
    Library = "Library"
    LibraryChests = "Library Chests"
    LibraryPopup = "Library Popup"
    MinnieEscort = "Minnie Escort"
    CornerstoneHill = "Cornerstone Hill"
    CornerstoneHillChests = "Cornerstone Hill Chests"
    Pier = "Pier"
    PierChests = "Pier Chests"
    Waterway = "Waterway"
    WaterwayChests = "Waterway Chests"
    WindowsPopup = "Windows Popup"
    BoatPete = "Boat Pete"
    FuturePete = "Future Pete"
    WisdomPopup = "Wisdom Popup"
    Marluxia = "AS Marluxia"
    DataMarluxia = "Data Marluxia"
    LingeringWill = "Lingering Will"


class CheckLocation(str, Enum):
    CourtyardMythrilShard = "DC Courtyard Mythril Shard"
    CourtyardStarRecipe = "DC Courtyard Star Recipe"
    CourtyardApBoost = "DC Courtyard AP Boost"
    CourtyardMythrilStone = "DC Courtyard Mythril Stone"
    CourtyardBlazingStone = "DC Courtyard Blazing Stone"
    CourtyardBlazingShard = "DC Courtyard Blazing Shard"
    CourtyardMythrilShard2 = "DC Courtyard Mythril Shard (2)"
    LibraryTornPages = "Library Torn Pages"
    DisneyCastleMap = "Disney Castle Map"
    MinnieEscort = "Minnie Escort"
    CornerstoneHillMap = "Cornerstone Hill Map"
    CornerstoneHillFrostShard = "Cornerstone Hill Frost Shard"
    PierMythrilShard = "Pier Mythril Shard"
    PierHiPotion = "Pier Hi-Potion"
    WaterwayMythrilStone = "Waterway Mythril Stone"
    WaterwayApBoost = "Waterway AP Boost"
    WaterwayFrostStone = "Waterway Frost Stone"
    WindowOfTimeMap = "Window of Time Map"
    BoatPete = "Boat Pete"
    FuturePete = "Future Pete"
    Monochrome = "Monochrome"
    WisdomForm = "Wisdom Form"
    MarluxiaBonus = "Marluxia Bonus"
    MarluxiaEternalBlossom = "Marluxia (AS) Eternal Blossom"
    DataMarluxiaLostIllusion = "Marluxia (Data) Lost Illusion"
    LingeringWillBonus = "Lingering Will Bonus"
    LingeringWillProofOfConnection = "Lingering Will Proof of Connection"
    LingeringWillManifestIllusion = "Lingering Will Manifest Illusion"

class DCLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_dc_keyblade(inv)
        self.logic[NodeId.DisneyCastleCourtyard][NodeId.DisneyCastleCourtyardChests] = keyblade_lambda
        self.logic[NodeId.Library][NodeId.LibraryChests] = keyblade_lambda
        self.logic[NodeId.CornerstoneHill][NodeId.CornerstoneHillChests] = keyblade_lambda
        self.logic[NodeId.Pier][NodeId.PierChests] = keyblade_lambda
        self.logic[NodeId.Waterway][NodeId.WaterwayChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.DisneyCastleCourtyard] = ItemPlacementHelpers.dc1_check
            self.logic[NodeId.MinnieEscort][NodeId.CornerstoneHill] = ItemPlacementHelpers.dc2_check
        else:
            self.logic[START_NODE][NodeId.CornerstoneHill] = ItemPlacementHelpers.dc1_check
            self.logic[NodeId.Marluxia][NodeId.DisneyCastleCourtyard] = ItemPlacementHelpers.dc2_check

        self.logic[NodeId.WisdomPopup][NodeId.LingeringWill] = ItemPlacementHelpers.need_proof_connection
        

def make_graph(graph: LocationGraphBuilder):
    dc = locationType.DC
    dc_logic = DCLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(dc_logic)

    courtyard_chests = graph.add_location(NodeId.DisneyCastleCourtyardChests, [
        chest(16, CheckLocation.CourtyardMythrilShard, dc),
        chest(17, CheckLocation.CourtyardStarRecipe, dc),
        chest(18, CheckLocation.CourtyardApBoost, dc),
        chest(92, CheckLocation.CourtyardMythrilStone, dc),
        chest(93, CheckLocation.CourtyardBlazingStone, dc),
        chest(247, CheckLocation.CourtyardBlazingShard, dc),
        chest(248, CheckLocation.CourtyardMythrilShard2, dc),
    ])
    courtyard = graph.add_location(NodeId.DisneyCastleCourtyard, [])
    library_chests = graph.add_location(NodeId.LibraryChests, [
        chest(91, CheckLocation.LibraryTornPages, dc, vanilla=misc.TornPages),
    ])
    library = graph.add_location(NodeId.Library, [])
    library_popup = graph.add_location(NodeId.LibraryPopup, [
        popup(332, CheckLocation.DisneyCastleMap, dc),
    ])
    minnie_escort = graph.add_location(NodeId.MinnieEscort, [
        hybrid_bonus(38, CheckLocation.MinnieEscort, dc, vanilla=ability.AutoSummon),
    ])
    cornerstone_hill_chests = graph.add_location(NodeId.CornerstoneHillChests, [
        chest(79, CheckLocation.CornerstoneHillMap, dc),
        chest(12, CheckLocation.CornerstoneHillFrostShard, dc),
    ])
    cornerstone_hill = graph.add_location(NodeId.CornerstoneHill, [])
    pier_chests = graph.add_location(NodeId.PierChests, [
        chest(81, CheckLocation.PierMythrilShard, dc),
        chest(82, CheckLocation.PierHiPotion, dc),
    ])
    pier = graph.add_location(NodeId.Pier, [])
    waterway_chests = graph.add_location(NodeId.WaterwayChests, [
        chest(83, CheckLocation.WaterwayMythrilStone, dc),
        chest(84, CheckLocation.WaterwayApBoost, dc),
        chest(85, CheckLocation.WaterwayFrostStone, dc),
    ])
    waterway = graph.add_location(NodeId.Waterway, [])
    windows_popup = graph.add_location(NodeId.WindowsPopup, [
        popup(368, CheckLocation.WindowOfTimeMap, dc),
    ])
    boat_pete = graph.add_location(NodeId.BoatPete, [
        item_bonus(16, CheckLocation.BoatPete, dc, vanilla=ability.DodgeSlash),
    ])
    future_pete = graph.add_location(NodeId.FuturePete, [
        hybrid_bonus(17, CheckLocation.FuturePete, dc, vanilla=magic.Reflect),
        popup(261, CheckLocation.Monochrome, dc, vanilla=keyblade.Monochrome),
    ])
    wisdom_popup = graph.add_location(NodeId.WisdomPopup, [
        popup(262, CheckLocation.WisdomForm, dc, vanilla=form.WisdomForm),
    ])
    marluxia = graph.add_location(NodeId.Marluxia, [
        stat_bonus(67, CheckLocation.MarluxiaBonus, [dc, locationType.AS]),
        popup(548, CheckLocation.MarluxiaEternalBlossom, [dc, locationType.AS]),
    ])
    data_marluxia = graph.add_location(NodeId.DataMarluxia, [
        popup(553, CheckLocation.DataMarluxiaLostIllusion, [dc, locationType.DataOrg]),
    ])
    lingering_will = graph.add_location(NodeId.LingeringWill, [
        stat_bonus(70, CheckLocation.LingeringWillBonus, [dc, locationType.LW],
                   invalid_checks=[itemType.PROOF_OF_CONNECTION]),
        popup(587, CheckLocation.LingeringWillProofOfConnection, [dc, locationType.LW],
              invalid_checks=[itemType.PROOF_OF_CONNECTION]),
        popup(591, CheckLocation.LingeringWillManifestIllusion, [dc, locationType.LW],
              invalid_checks=[itemType.PROOF_OF_CONNECTION]),
    ])

    graph.register_superboss(data_marluxia)
    graph.register_superboss(lingering_will)
    graph.register_last_story_boss(wisdom_popup)
    
    graph.add_edge(courtyard, courtyard_chests)
    graph.add_edge(library, library_chests)
    graph.add_edge(cornerstone_hill, cornerstone_hill_chests)
    graph.add_edge(pier, pier_chests)
    graph.add_edge(waterway, waterway_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, courtyard)
        graph.add_edge(courtyard, library)
        graph.add_edge(library, library_popup)
        graph.add_edge(library_popup, minnie_escort)
        graph.add_edge(minnie_escort, cornerstone_hill)
        graph.add_edge(cornerstone_hill, pier)
        graph.add_edge(pier, waterway, RequirementEdge(battle=True))
        graph.add_edge(waterway, windows_popup, RequirementEdge(battle=True))
        graph.add_edge(windows_popup, boat_pete, RequirementEdge(battle=True))
        graph.add_edge(boat_pete, future_pete, RequirementEdge(battle=True))
        graph.add_edge(future_pete, wisdom_popup)
        graph.add_edge(wisdom_popup, marluxia, RequirementEdge(battle=True))
        graph.add_edge(marluxia, data_marluxia)
        graph.add_edge(wisdom_popup, lingering_will, RequirementEdge(battle=True))

        graph.register_first_boss(minnie_escort)
        graph.register_superboss(marluxia)
    else:
        graph.add_edge(START_NODE, cornerstone_hill)
        graph.add_edge(cornerstone_hill, pier)
        graph.add_edge(pier, waterway, RequirementEdge(battle=True))
        graph.add_edge(waterway, windows_popup, RequirementEdge(battle=True))
        graph.add_edge(windows_popup, boat_pete, RequirementEdge(battle=True))
        graph.add_edge(boat_pete, future_pete, RequirementEdge(battle=True))
        graph.add_edge(future_pete, marluxia, RequirementEdge(battle=True))
        graph.add_edge(marluxia, courtyard)
        graph.add_edge(courtyard, library)
        graph.add_edge(library, library_popup)
        graph.add_edge(library_popup, minnie_escort)
        graph.add_edge(minnie_escort, wisdom_popup)
        graph.add_edge(wisdom_popup, data_marluxia, RequirementEdge(battle=True))
        graph.add_edge(wisdom_popup, lingering_will,RequirementEdge(battle=True))
        graph.register_first_boss(future_pete)



from enum import Enum

from List.configDict import locationType
from List.inventory import magic, keyblade
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    PoohsHowse = "Pooh's Howse"
    PoohsHowseChests = "Pooh's Howse Chests"
    PigletsHowse = "Piglet's Howse"
    PigletsHowseChests = "Piglet's Howse Chests"
    RabbitsHowse = "Rabbit's Howse"
    RabbitsHowseChests = "Rabbit's Howse Chests"
    KangasHowse = "Kanga's Howse"
    KangasHowseChests = "Kanga's Howse Chests"
    SpookyCave = "Spooky Cave"
    SpookyCaveChests = "Spooky Cave Chests"
    StarryHill = "Starry Hill"
    StarryHillChests = "Starry Hill Chests"
    StarryHillPopups = "Starry Hill Popups"


class CheckLocation(str, Enum):
    PoohsHowseHundredAcreWoodMap = "Pooh's House 100 Acre Wood Map"
    PoohsHowseApBoost = "Pooh's House AP Boost"
    PoohsHowseMythrilStone = "Pooh's House Mythril Stone"
    PigletsHowseDefenseBoost = "Piglet's House Defense Boost"
    PigletsHowseApBoost = "Piglet's House AP Boost"
    PigletsHowseMythrilGem = "Piglet's House Mythril Gem"
    RabbitsHowseDrawRing = "Rabbit's House Draw Ring"
    RabbitsHowseMythrilCrystal = "Rabbit's House Mythril Crystal"
    RabbitsHowseApBoost = "Rabbit's House AP Boost"
    KangasHowseMagicBoost = "Kanga's House Magic Boost"
    KangasHowseApBoost = "Kanga's House AP Boost"
    KangasHowseOrichalcum = "Kanga's House Orichalcum"
    SpookyCaveMyhtrilGem = "Spooky Cave Mythril Gem"
    SpookyCaveApBoost = "Spooky Cave AP Boost"
    SpookyCaveOrichalcum = "Spooky Cave Orichalcum"
    SpookyCaveGuardRecipe = "Spooky Cave Guard Recipe"
    SpookyCaveMythrilCrystal = "Spooky Cave Mythril Crystal"
    SpookyCaveApBoost2 = "Spooky Cave AP Boost (2)"
    SweetMemories = "Sweet Memories"
    SpookyCaveMap = "Spooky Cave Map"
    StarryHillCosmicRing = "Starry Hill Cosmic Ring"
    StarryHillStyleRecipe = "Starry Hill Style Recipe"
    StarryHillCureElement = "Starry Hill Cure Element"
    StarryHillOrichalcumPlus = "Starry Hill Orichalcum+"

class HAWLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando, keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_haw_keyblade(inv)
        self.logic[NodeId.PoohsHowse][NodeId.PoohsHowseChests] = keyblade_lambda
        self.logic[NodeId.PigletsHowse][NodeId.PigletsHowseChests] = keyblade_lambda
        self.logic[NodeId.RabbitsHowse][NodeId.RabbitsHowseChests] = keyblade_lambda
        self.logic[NodeId.KangasHowse][NodeId.KangasHowseChests] = keyblade_lambda
        self.logic[NodeId.SpookyCave][NodeId.SpookyCaveChests] = keyblade_lambda
        self.logic[NodeId.StarryHill][NodeId.StarryHillChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[NodeId.PoohsHowse][NodeId.PigletsHowse] = ItemPlacementHelpers.need_torn_pages(1)
            self.logic[NodeId.PigletsHowse][NodeId.RabbitsHowse] = ItemPlacementHelpers.need_torn_pages(2)
            self.logic[NodeId.RabbitsHowse][NodeId.KangasHowse] = ItemPlacementHelpers.need_torn_pages(3)
            self.logic[NodeId.KangasHowse][NodeId.SpookyCave] = ItemPlacementHelpers.need_torn_pages(4)
            self.logic[NodeId.SpookyCave][NodeId.StarryHill] = ItemPlacementHelpers.need_torn_pages(5)
        else:
            self.logic[START_NODE][NodeId.StarryHill] = keyblade_lambda # need to do this because pages can lock themselves in Reverse Rando
            self.logic[NodeId.StarryHill][NodeId.SpookyCave] = ItemPlacementHelpers.need_torn_pages(1)
            self.logic[NodeId.SpookyCave][NodeId.KangasHowse] = ItemPlacementHelpers.need_torn_pages(2)
            self.logic[NodeId.KangasHowse][NodeId.RabbitsHowse] = ItemPlacementHelpers.need_torn_pages(3)
            self.logic[NodeId.RabbitsHowse][NodeId.PigletsHowse] = ItemPlacementHelpers.need_torn_pages(4)
            self.logic[NodeId.PigletsHowse][NodeId.PoohsHowse] = ItemPlacementHelpers.need_torn_pages(5)
            

def make_graph(graph: LocationGraphBuilder):
    haw = locationType.HUNDREDAW
    haw_logic = HAWLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(haw_logic)

    poohs_howse_chests = graph.add_location(NodeId.PoohsHowseChests, [
        chest(313, CheckLocation.PoohsHowseHundredAcreWoodMap, haw),
        chest(97, CheckLocation.PoohsHowseApBoost, haw),
        chest(98, CheckLocation.PoohsHowseMythrilStone, haw),
    ])
    poohs_howse = graph.add_location(NodeId.PoohsHowse, [])
    piglets_howse_chests = graph.add_location(NodeId.PigletsHowseChests, [
        chest(105, CheckLocation.PigletsHowseDefenseBoost, haw),
        chest(103, CheckLocation.PigletsHowseApBoost, haw),
        chest(104, CheckLocation.PigletsHowseMythrilGem, haw),
    ])
    piglets_howse = graph.add_location(NodeId.PigletsHowse, [])
    rabbits_howse_chests = graph.add_location(NodeId.RabbitsHowseChests, [
        chest(314, CheckLocation.RabbitsHowseDrawRing, haw),
        chest(100, CheckLocation.RabbitsHowseMythrilCrystal, haw),
        chest(101, CheckLocation.RabbitsHowseApBoost, haw),
    ])
    rabbits_howse = graph.add_location(NodeId.RabbitsHowse, [])
    kangas_howse_chests = graph.add_location(NodeId.KangasHowseChests, [
        chest(108, CheckLocation.KangasHowseMagicBoost, haw),
        chest(106, CheckLocation.KangasHowseApBoost, haw),
        chest(107, CheckLocation.KangasHowseOrichalcum, haw),
    ])
    kangas_howse = graph.add_location(NodeId.KangasHowse, [])
    spooky_cave_chests = graph.add_location(NodeId.SpookyCaveChests, [
        chest(110, CheckLocation.SpookyCaveMyhtrilGem, haw),
        chest(111, CheckLocation.SpookyCaveApBoost, haw),
        chest(112, CheckLocation.SpookyCaveOrichalcum, haw),
        chest(113, CheckLocation.SpookyCaveGuardRecipe, haw),
        chest(115, CheckLocation.SpookyCaveMythrilCrystal, haw),
        chest(116, CheckLocation.SpookyCaveApBoost2, haw),
    ])
    spooky_cave = graph.add_location(NodeId.SpookyCave, [
        popup(284, CheckLocation.SweetMemories, haw, vanilla=keyblade.SweetMemories),
        popup(485, CheckLocation.SpookyCaveMap, haw),
    ])
    starry_hill_chests = graph.add_location(NodeId.StarryHillChests, [
        chest(312, CheckLocation.StarryHillCosmicRing, haw),
        chest(94, CheckLocation.StarryHillStyleRecipe, haw),
    ])
    starry_hill = graph.add_location(NodeId.StarryHill, [])
    
    starry_hill_popups = graph.add_location(NodeId.StarryHillPopups, [
        popup(285, CheckLocation.StarryHillCureElement, haw, vanilla=magic.Cure),
        popup(539, CheckLocation.StarryHillOrichalcumPlus, haw),
    ])

    graph.add_edge(poohs_howse, poohs_howse_chests)
    graph.add_edge(piglets_howse, piglets_howse_chests)
    graph.add_edge(rabbits_howse, rabbits_howse_chests)
    graph.add_edge(kangas_howse, kangas_howse_chests)
    graph.add_edge(spooky_cave, spooky_cave_chests)
    graph.add_edge(starry_hill, starry_hill_chests)
    graph.add_edge(starry_hill, starry_hill_popups)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, poohs_howse)
        graph.add_edge(poohs_howse, piglets_howse)
        graph.add_edge(piglets_howse, rabbits_howse)
        graph.add_edge(rabbits_howse, kangas_howse)
        graph.add_edge(kangas_howse, spooky_cave)
        graph.add_edge(spooky_cave, starry_hill)
    else:
        graph.add_edge(START_NODE, starry_hill)
        graph.add_edge(starry_hill, spooky_cave)
        graph.add_edge(spooky_cave, kangas_howse)
        graph.add_edge(kangas_howse, rabbits_howse)
        graph.add_edge(rabbits_howse, piglets_howse)
        graph.add_edge(piglets_howse, poohs_howse)


def yeet_the_bear_location_names() -> list[str]:
    """ The names of the "yeet the bear" popup locations. """
    return [CheckLocation.StarryHillCureElement, CheckLocation.StarryHillOrichalcumPlus]



from Class.itemClass import KH2Item
from List.inventory import ability, form, growth, magic, misc, proof, storyunlock, summon, synth, keyblade
from List.inventory.form import DriveForm
from List.inventory.growth import GrowthType
from List.inventory.item import InventoryItem
from List.location.graph import RequirementFunction


class ItemPlacementHelpers:

    @staticmethod
    def need_fire_blizzard_thunder(inventory: list[int]) -> bool:
        return magic.Fire.id in inventory and magic.Blizzard.id in inventory and magic.Thunder.id in inventory

    @staticmethod
    def need_1_magnet(inventory: list[int]) -> bool:
        return inventory.count(magic.Magnet.id) >= 1

    @staticmethod
    def need_2_magnet(inventory: list[int]) -> bool:
        return inventory.count(magic.Magnet.id) >= 2

    @staticmethod
    def need_3_thunders(inventory: list[int]) -> bool:
        return inventory.count(magic.Thunder.id) == 3

    @staticmethod
    def count_growth(inventory: list[int], growth_type: GrowthType) -> int:
        ids = [item.id for item in growth.all_growth() if item.growth_type == growth_type]
        count = 0
        for item in inventory:
            if item in ids:
                count += 1
        return count

    @staticmethod
    def need_growths(inventory: list[int]) -> bool:
        return ItemPlacementHelpers.count_growth(inventory, GrowthType.HIGH_JUMP) >= 3 \
            and ItemPlacementHelpers.count_growth(inventory, GrowthType.QUICK_RUN) >= 3 \
            and ItemPlacementHelpers.count_growth(inventory, GrowthType.AERIAL_DODGE) >= 3 \
            and ItemPlacementHelpers.count_growth(inventory, GrowthType.GLIDE) >= 3

    @staticmethod
    def need_proof_connection(inventory: list[int]) -> bool:
        return proof.ProofOfConnection.id in inventory

    @staticmethod
    def need_proof_peace(inventory: list[int]) -> bool:
        return proof.ProofOfPeace.id in inventory

    @staticmethod
    def has_valor_form(inventory: list[int]) -> bool:
        return form.ValorForm.id in inventory

    @staticmethod
    def has_wisdom_form(inventory: list[int]) -> bool:
        return form.WisdomForm.id in inventory

    @staticmethod
    def has_limit_form(inventory: list[int]) -> bool:
        return form.LimitForm.id in inventory

    @staticmethod
    def has_master_form(inventory: list[int]) -> bool:
        return form.MasterForm.id in inventory

    @staticmethod
    def has_final_form(inventory: list[int]) -> bool:
        return form.FinalForm.id in inventory
    
    @staticmethod
    def can_level_valor(inventory: list[int]) -> bool:
        return True
    @staticmethod
    def can_level_wisdom(inventory: list[int]) -> bool:
        return True
    @staticmethod
    def can_level_limit(inventory: list[int]) -> bool:
        return True
    @staticmethod
    def can_level_master(inventory: list[int]) -> bool:
        return ItemPlacementHelpers.ht1_check(inventory) \
              or ItemPlacementHelpers.hb2_check(inventory)
    @staticmethod
    def can_level_final(inventory: list[int]) -> bool:
        return ItemPlacementHelpers.twtnw_roxas_check(inventory) \
              or ItemPlacementHelpers.pr2_check(inventory) \
              or ItemPlacementHelpers.bc2_check(inventory) \
              or ItemPlacementHelpers.lod2_check(inventory) \
              or ItemPlacementHelpers.tt3_check(inventory)

    @staticmethod
    def has_auto_valor(inventory: list[int]) -> bool:
        return ability.AutoValor.id in inventory

    @staticmethod
    def has_auto_wisdom(inventory: list[int]) -> bool:
        return ability.AutoWisdom.id in inventory

    @staticmethod
    def has_auto_limit(inventory: list[int]) -> bool:
        return ability.AutoLimitForm.id in inventory

    @staticmethod
    def has_auto_master(inventory: list[int]) -> bool:
        return ability.AutoMaster.id in inventory

    @staticmethod
    def has_auto_final(inventory: list[int]) -> bool:
        return ability.AutoFinal.id in inventory

    @staticmethod
    def count_forms(inventory: list[int]) -> int:
        count = 0
        if ItemPlacementHelpers.has_valor_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_wisdom_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_limit_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_master_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_final_form(inventory):
            count += 1
        return count

    @staticmethod
    def need_forms(inventory: list[int]) -> bool:
        return ItemPlacementHelpers.count_forms(inventory) == 5

    @staticmethod
    def need_summons(inventory: list[int]) -> bool:
        return summon.LampCharm.id in inventory \
            and summon.FeatherCharm.id in inventory \
            and summon.UkuleleCharm.id in inventory \
            and summon.BaseballCharm.id in inventory

    @staticmethod
    def count_pages(inventory: list[int]) -> int:
        return inventory.count(misc.TornPages.id)

    @staticmethod
    def need_torn_pages(count: int) -> RequirementFunction:
        return lambda inventory: ItemPlacementHelpers.count_pages(inventory) >= count

    @staticmethod
    def need_proofs(inventory: list[int]) -> bool:
        return proof.ProofOfConnection.id in inventory \
            and proof.ProofOfNonexistence.id in inventory \
            and proof.ProofOfPeace.id in inventory
    
    @staticmethod
    def need_promise_charm(inventory: list[int]) -> bool:
        return misc.PromiseCharm.id in inventory
    
    @staticmethod
    def make_need_objectives_lambda(num_objectives_needed: int):
        return lambda inventory : inventory.count(misc.ObjectiveItem.id) >= num_objectives_needed        

    @staticmethod
    def make_form_lambda(drive_form: DriveForm, form_level: int) -> RequirementFunction:
        if drive_form == form.ValorForm:
            return lambda inventory: ItemPlacementHelpers.has_valor_form(inventory) \
                                     and ItemPlacementHelpers.can_level_valor(inventory) \
                                     and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
        if drive_form == form.WisdomForm:
            return lambda inventory: ItemPlacementHelpers.has_wisdom_form(inventory) \
                                     and ItemPlacementHelpers.can_level_wisdom(inventory) \
                                     and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
        if drive_form == form.LimitForm:
            return lambda inventory: ItemPlacementHelpers.has_limit_form(inventory) \
                                     and ItemPlacementHelpers.can_level_limit(inventory) \
                                     and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
        if drive_form == form.MasterForm:
            return lambda inventory: ItemPlacementHelpers.has_master_form(inventory) \
                                     and ItemPlacementHelpers.can_level_master(inventory) \
                                     and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
        if drive_form == form.FinalForm:
            return lambda inventory: ItemPlacementHelpers.has_final_form(inventory) \
                                     and ItemPlacementHelpers.can_level_final(inventory) \
                                     and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
        return lambda inventory: False

    @staticmethod
    def make_form_lambda_nightmare(drive_form: DriveForm, form_level: int) -> RequirementFunction:
        def count_auto_forms(inventory: list[int]) -> int:
            count = 0
            if ItemPlacementHelpers.has_auto_valor(inventory):
                count += 1
            if ItemPlacementHelpers.has_auto_wisdom(inventory):
                count += 1
            if ItemPlacementHelpers.has_auto_limit(inventory):
                count += 1
            if ItemPlacementHelpers.has_auto_master(inventory):
                count += 1
            return count

        def final_possible_but_not_obtained(inventory: list[int]) -> bool:
            return (ItemPlacementHelpers.has_valor_form(inventory)
                    or ItemPlacementHelpers.has_wisdom_form(inventory)
                    or ItemPlacementHelpers.has_limit_form(inventory)
                    or ItemPlacementHelpers.has_master_form(inventory)
                    or count_auto_forms(inventory) >= 1) \
                and not ItemPlacementHelpers.has_final_form(inventory)

        def max_form_level(inventory: list[int]) -> int:
            base_count = ItemPlacementHelpers.count_forms(inventory) + 2
            if final_possible_but_not_obtained(inventory):
                return base_count + 1
            else:
                return base_count

        if drive_form == form.ValorForm:
            return lambda inventory: (ItemPlacementHelpers.has_valor_form(inventory)
                                      or ItemPlacementHelpers.has_auto_valor(inventory)) \
                                     and ItemPlacementHelpers.can_level_valor(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.WisdomForm:
            return lambda inventory: (ItemPlacementHelpers.has_wisdom_form(inventory)
                                      or ItemPlacementHelpers.has_auto_wisdom(inventory)) \
                                     and ItemPlacementHelpers.can_level_wisdom(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.LimitForm:
            return lambda inventory: (ItemPlacementHelpers.has_limit_form(inventory)
                                      or ItemPlacementHelpers.has_auto_limit(inventory)) \
                                     and ItemPlacementHelpers.can_level_limit(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.MasterForm:
            return lambda inventory: (ItemPlacementHelpers.has_master_form(inventory)
                                      or ItemPlacementHelpers.has_auto_master(inventory)) \
                                     and ItemPlacementHelpers.can_level_master(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.FinalForm:
            def have_final_form(inventory: list[int]) -> bool:
                return ItemPlacementHelpers.has_final_form(inventory) or final_possible_but_not_obtained(inventory)

            if form_level == 2:
                return lambda inventory: (have_final_form(inventory) or ItemPlacementHelpers.has_auto_final(inventory)) \
                                     and ItemPlacementHelpers.can_level_final(inventory)
            else:
                return lambda inventory: have_final_form(inventory) \
                                     and max_form_level(inventory) >= form_level \
                                     and ItemPlacementHelpers.can_level_final(inventory)

        return lambda inventory: False


    @staticmethod
    def make_form_lambda_nightmare_no_anti(drive_form: DriveForm, form_level: int) -> RequirementFunction:
        def max_form_level(inventory: list[int]) -> int:
            base_count = ItemPlacementHelpers.count_forms(inventory) + 2
            return base_count

        if drive_form == form.ValorForm:
            return lambda inventory: (ItemPlacementHelpers.has_valor_form(inventory)
                                      or ItemPlacementHelpers.has_auto_valor(inventory)) \
                                     and ItemPlacementHelpers.can_level_valor(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.WisdomForm:
            return lambda inventory: (ItemPlacementHelpers.has_wisdom_form(inventory)
                                      or ItemPlacementHelpers.has_auto_wisdom(inventory)) \
                                     and ItemPlacementHelpers.can_level_wisdom(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.LimitForm:
            return lambda inventory: (ItemPlacementHelpers.has_limit_form(inventory)
                                      or ItemPlacementHelpers.has_auto_limit(inventory)) \
                                     and ItemPlacementHelpers.can_level_limit(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.MasterForm:
            return lambda inventory: (ItemPlacementHelpers.has_master_form(inventory)
                                      or ItemPlacementHelpers.has_auto_master(inventory)) \
                                     and ItemPlacementHelpers.can_level_master(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.FinalForm:
            return lambda inventory: (ItemPlacementHelpers.has_final_form(inventory)
                                      or ItemPlacementHelpers.has_auto_final(inventory)) \
                                     and ItemPlacementHelpers.can_level_final(inventory) \
                                     and max_form_level(inventory) >= form_level

        return lambda inventory: False

    @staticmethod
    def make_form_lambda_nightmare_no_final(drive_form: DriveForm, form_level: int) -> RequirementFunction:
        def final_possible_but_not_obtained(_: list[int]) -> bool:
            return False

        def max_form_level(inventory: list[int]) -> int:
            base_count = ItemPlacementHelpers.count_forms(inventory) + 2
            if final_possible_but_not_obtained(inventory):
                return base_count + 1
            else:
                return base_count

        if drive_form == form.ValorForm:
            return lambda inventory: (ItemPlacementHelpers.has_valor_form(inventory)
                                      or ItemPlacementHelpers.has_auto_valor(inventory)) \
                                     and ItemPlacementHelpers.can_level_valor(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.WisdomForm:
            return lambda inventory: (ItemPlacementHelpers.has_wisdom_form(inventory)
                                      or ItemPlacementHelpers.has_auto_wisdom(inventory)) \
                                     and ItemPlacementHelpers.can_level_wisdom(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.LimitForm:
            return lambda inventory: (ItemPlacementHelpers.has_limit_form(inventory)
                                      or ItemPlacementHelpers.has_auto_limit(inventory)) \
                                     and ItemPlacementHelpers.can_level_limit(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.MasterForm:
            return lambda inventory: (ItemPlacementHelpers.has_master_form(inventory)
                                      or ItemPlacementHelpers.has_auto_master(inventory)) \
                                     and ItemPlacementHelpers.can_level_master(inventory) \
                                     and max_form_level(inventory) >= form_level
        if drive_form == form.FinalForm:
            def have_final_form(inventory: list[int]) -> bool:
                return final_possible_but_not_obtained(inventory) or ItemPlacementHelpers.has_final_form(inventory)

            if form_level == 2:
                return lambda inventory: (have_final_form(inventory) or ItemPlacementHelpers.has_auto_final(inventory)) \
                                     and ItemPlacementHelpers.can_level_final(inventory)
            else:
                return lambda inventory: have_final_form(inventory) and max_form_level(inventory) >= form_level \
                                     and ItemPlacementHelpers.can_level_final(inventory)

        return lambda inventory: False

    @staticmethod
    def get_number_visit_unlocks(inventory: list[int]) -> int:
        visit_unlock_ids = [u.id for u in storyunlock.all_story_unlocks()]
        running_total = 0
        for id in visit_unlock_ids:
            running_total+=inventory.count(id)
        return running_total

    @staticmethod
    def make_level_group_check(group_index) -> RequirementFunction:
        thresholds = [1,3,6,9,12,15,18,21,24] # modified from AP so that max level could in theory have a visit unlock
        if group_index<len(thresholds):
            return lambda inv: ItemPlacementHelpers.get_number_visit_unlocks(inv) >= thresholds[group_index]
        else:
            return lambda inv: ItemPlacementHelpers.get_number_visit_unlocks(inv) >= thresholds[-1]


    @staticmethod
    def stt_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.NaminesSketches.id)==1
    
    @staticmethod
    def dc1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.RoyalSummons.id)>=1

    @staticmethod
    def dc2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.RoyalSummons.id)==2

    @staticmethod
    def oc1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BattlefieldsOfWar.id)>=1

    @staticmethod
    def oc2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BattlefieldsOfWar.id)==2

    @staticmethod
    def lod1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SwordOfTheAncestor.id)>=1
    
    @staticmethod
    def lod2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SwordOfTheAncestor.id)==2

    @staticmethod
    def bc1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BeastsClaw.id)>=1
    
    @staticmethod
    def bc2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BeastsClaw.id)==2

    @staticmethod
    def ht1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BoneFist.id)>=1
    
    @staticmethod
    def ht2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BoneFist.id)==2

    @staticmethod
    def pl1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.ProudFang.id)>=1
    
    @staticmethod
    def pl2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.ProudFang.id)==2
    
    @staticmethod
    def pr1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SkillAndCrossbones.id)>=1

    @staticmethod
    def pr2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SkillAndCrossbones.id)==2

    @staticmethod
    def ag1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.Scimitar.id)>=1

    @staticmethod
    def ag2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.Scimitar.id)==2

    @staticmethod
    def twtnw_roxas_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.WayToTheDawn.id) >= 1 and ItemPlacementHelpers.get_number_visit_unlocks(inventory) > 10
    
    @staticmethod
    def twtnw_post_saix_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.WayToTheDawn.id) == 2

    @staticmethod
    def sp1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IdentityDisk.id)>=1
    
    @staticmethod
    def sp2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IdentityDisk.id) == 2

    @staticmethod
    def tt1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IceCream.id) >= 1

    @staticmethod
    def tt2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IceCream.id) >= 2

    @staticmethod
    def tt3_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IceCream.id) == 3

    @staticmethod
    def hb1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.MembershipCard.id) >= 1

    @staticmethod
    def hb2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.MembershipCard.id) == 2

    @staticmethod
    def make_synth_requirement(synth_item: KH2Item) -> RequirementFunction:
        # making lambdas for the different synth items
        default_access_lambda: RequirementFunction = lambda inventory: True
        synth_req_map: dict[InventoryItem, RequirementFunction] = {
            synth.BlazingShard: lambda inv: ItemPlacementHelpers.dc2_check(inv) \
                                               or ItemPlacementHelpers.bc2_check(inv) \
                                               or ItemPlacementHelpers.oc2_check(inv),

            synth.BlazingStone: lambda inv: ItemPlacementHelpers.lod2_check(inv) \
                                               or ItemPlacementHelpers.sp2_check(inv) \
                                               or ItemPlacementHelpers.oc2_check(inv),

            synth.BlazingGem: ItemPlacementHelpers.ag1_check,

            synth.BlazingCrystal: lambda inv: ItemPlacementHelpers.oc2_check(inv) \
                                                 or ItemPlacementHelpers.hb2_check(inv) \
                                                 or ItemPlacementHelpers.bc2_check(inv) \
                                                 or ItemPlacementHelpers.ag2_check(inv),

            # shadows are guaranteed anytime in DC/TR
            synth.DarkShard: ItemPlacementHelpers.dc1_check,

            # LoD1 and 2 both have them
            synth.DarkStone: ItemPlacementHelpers.lod1_check,

            # BC1 and 2 have them
            synth.DarkGem: ItemPlacementHelpers.bc1_check,

            # PR1 has them, but could be locked out
            synth.DarkCrystal: lambda inv: ItemPlacementHelpers.lod2_check(inv),

            # TWTNW has them post Xemnas
            synth.DenseShard: lambda inv: ItemPlacementHelpers.twtnw_post_saix_check(inv) \
                                            or ItemPlacementHelpers.bc2_check(inv) \
                                            or ItemPlacementHelpers.tt3_check(inv),

            # TWTNW has them post Xemnas
            synth.DenseStone: lambda inv: ItemPlacementHelpers.lod2_check(inv) \
                                            or ItemPlacementHelpers.twtnw_roxas_check(inv),

            # TWTNW has them post Xemnas
            synth.DenseGem: lambda inv: ItemPlacementHelpers.tt3_check(inv) \
                                            or ItemPlacementHelpers.stt_check(inv) \
                                            or ItemPlacementHelpers.twtnw_roxas_check(inv),

            # TWTNW has them post Xemnas
            synth.DenseCrystal:  lambda inv: ItemPlacementHelpers.tt3_check(inv) \
                                            or ItemPlacementHelpers.twtnw_roxas_check(inv) \
                                            or ItemPlacementHelpers.pl2_check(inv),

            # BC ones can disappear
            synth.FrostShard: lambda inv: ItemPlacementHelpers.lod2_check(inv) \
                                             or ItemPlacementHelpers.ag2_check(inv) \
                                             or ItemPlacementHelpers.pr2_check(inv),

            # DC/TR
            synth.FrostStone: ItemPlacementHelpers.dc2_check,

            # AG1 and 2 both have them
            synth.FrostGem: ItemPlacementHelpers.ag1_check,

            # PL1 and 2 both have them
            synth.FrostCrystal: ItemPlacementHelpers.pl1_check,

            # LoD1 and 2 both have them
            synth.LightningShard: lambda inv: ItemPlacementHelpers.lod1_check(inv) \
                                                or ItemPlacementHelpers.dc1_check(inv) \
                                                or ItemPlacementHelpers.pl2_check(inv) \
                                                or ItemPlacementHelpers.pr2_check(inv) \
                                                or ItemPlacementHelpers.ag2_check(inv),

            # HT and SP ones can become unavailable
            synth.LightningStone: lambda inv: ItemPlacementHelpers.oc2_check(inv) \
                                                 or ItemPlacementHelpers.lod2_check(inv) \
                                                 or ItemPlacementHelpers.ag2_check(inv),

            synth.LightningGem: lambda inv: ItemPlacementHelpers.hb2_check(inv) \
                                               or ItemPlacementHelpers.bc2_check(inv) \
                                               or ItemPlacementHelpers.pr2_check(inv) \
                                               or ItemPlacementHelpers.ht2_check(inv) \
                                               or ItemPlacementHelpers.pl2_check(inv) \
                                               or ItemPlacementHelpers.sp2_check(inv),

            # SP1 and 2 both have them
            synth.LightningCrystal: ItemPlacementHelpers.sp1_check,

            # OC1 can be locked out
            synth.LucidShard: lambda inv: ItemPlacementHelpers.pr2_check(inv) \
                                             or ItemPlacementHelpers.ht2_check(inv),

            # HT1 and 2 have them
            synth.LucidStone: ItemPlacementHelpers.ht1_check,

            # SP1 and 2 both have them
            synth.LucidGem: lambda inv: ItemPlacementHelpers.hb2_check(inv) \
                                            or ItemPlacementHelpers.sp1_check(inv),

            synth.LucidCrystal: lambda inv: ItemPlacementHelpers.lod2_check(inv) \
                                               or ItemPlacementHelpers.bc2_check(inv) \
                                               or ItemPlacementHelpers.oc2_check(inv) \
                                               or ItemPlacementHelpers.pr2_check(inv) \
                                               or ItemPlacementHelpers.ag2_check(inv) \
                                               or ItemPlacementHelpers.ht2_check(inv) \
                                               or ItemPlacementHelpers.pl2_check(inv) \
                                               or ItemPlacementHelpers.hb2_check(inv),

            # BC1 and 2 both have them, also TR
            synth.PowerShard: lambda inv: ItemPlacementHelpers.bc1_check(inv)\
                                          or ItemPlacementHelpers.dc2_check(inv)\
                                          or ItemPlacementHelpers.ht2_check(inv),

            # AG1 and 2 both have them
            synth.PowerStone: ItemPlacementHelpers.ag1_check,

            # PL1 and 2 both have them
            synth.PowerGem: ItemPlacementHelpers.pl1_check,

            synth.PowerCrystal: lambda inv: ItemPlacementHelpers.hb2_check(inv) \
                                               or ItemPlacementHelpers.bc2_check(inv) \
                                               or ItemPlacementHelpers.oc2_check(inv) \
                                               or ItemPlacementHelpers.pr2_check(inv),

            # TWTNW always has them in Luxord and Saix rooms
            synth.TwilightShard: lambda inv: ItemPlacementHelpers.twtnw_roxas_check(inv)\
                                                or ItemPlacementHelpers.pr2_check(inv)\
                                                or ItemPlacementHelpers.tt3_check(inv)\
                                                or ItemPlacementHelpers.lod2_check(inv),

            # TWTNW always has them
            synth.TwilightStone: lambda inv: ItemPlacementHelpers.tt3_check(inv)\
                                                or ItemPlacementHelpers.twtnw_roxas_check(inv)\
                                                or ItemPlacementHelpers.hb2_check(inv),

            # TWTNW always has them
            synth.TwilightGem: lambda inv: ItemPlacementHelpers.twtnw_roxas_check(inv)\
                                            or ItemPlacementHelpers.tt3_check(inv),

            # TWTNW always has them
            synth.TwilightCrystal: ItemPlacementHelpers.twtnw_roxas_check,
        }

        item = synth_item.item
        if item in synth_req_map:
            return synth_req_map[item]

        # default returned lambda, doesn't allow for acquiring the item, failsafe
        return lambda inventory: False

    @staticmethod
    def need_stt_keyblade(inventory: list[int]) -> bool:
        return keyblade.BondOfFlame.id in inventory
    @staticmethod
    def need_tt_keyblade(inventory: list[int]) -> bool:
        return keyblade.Oathkeeper.id in inventory
    @staticmethod
    def need_hb_keyblade(inventory: list[int]) -> bool:
        return keyblade.SleepingLion.id in inventory
    @staticmethod
    def need_cor_keyblade(inventory: list[int]) -> bool:
        return keyblade.WinnersProof.id in inventory
    @staticmethod
    def need_lod_keyblade(inventory: list[int]) -> bool:
        return keyblade.HiddenDragon.id in inventory
    @staticmethod
    def need_bc_keyblade(inventory: list[int]) -> bool:
        return keyblade.RumblingRose.id in inventory
    @staticmethod
    def need_oc_keyblade(inventory: list[int]) -> bool:
        return keyblade.HerosCrest.id in inventory
    @staticmethod
    def need_dc_keyblade(inventory: list[int]) -> bool:
        return keyblade.Monochrome.id in inventory
    @staticmethod
    def need_pr_keyblade(inventory: list[int]) -> bool:
        return keyblade.FollowTheWind.id in inventory
    @staticmethod
    def need_ag_keyblade(inventory: list[int]) -> bool:
        return keyblade.WishingLamp.id in inventory
    @staticmethod
    def need_ht_keyblade(inventory: list[int]) -> bool:
        return keyblade.DecisivePumpkin.id in inventory
    @staticmethod
    def need_pl_keyblade(inventory: list[int]) -> bool:
        return keyblade.CircleOfLife.id in inventory
    @staticmethod
    def need_sp_keyblade(inventory: list[int]) -> bool:
        return keyblade.PhotonDebugger.id in inventory
    @staticmethod
    def need_twtnw_keyblade(inventory: list[int]) -> bool:
        return keyblade.TwoBecomeOne.id in inventory
    @staticmethod
    def need_haw_keyblade(inventory: list[int]) -> bool:
        return keyblade.SweetMemories.id in inventory




id_to_enemy_name = {}

id_to_enemy_name[1] = "Soldier"
id_to_enemy_name[2] = "Shadow"
id_to_enemy_name[3] = "Large Body"
id_to_enemy_name[4] = "Armored Knight"
id_to_enemy_name[5] = "Surveillance Robot"
id_to_enemy_name[6] = "Dusk"
id_to_enemy_name[7] = "Trick Ghost"
id_to_enemy_name[8] = "Rabid Dog"
id_to_enemy_name[9] = "Hook Bat"
id_to_enemy_name[10] = "Minute Bomb"
id_to_enemy_name[11] = "Assault Rider"
id_to_enemy_name[12] = "Hammer Frame"
id_to_enemy_name[13] = "Aeroplane"
id_to_enemy_name[14] = "OC Torches"
id_to_enemy_name[15] = "Samurai"
id_to_enemy_name[16] = "OC Bubbles"
id_to_enemy_name[17] = "Rapid Thruster"
id_to_enemy_name[18] = "Bolt Tower"
# id_to_enemy_name[19] = "" # mp only drops
id_to_enemy_name[22] = "Dragoon"
id_to_enemy_name[23] = "Assassin"
id_to_enemy_name[24] = "Sniper"
id_to_enemy_name[25] = "Dancer"
id_to_enemy_name[26] = "Berserker"
id_to_enemy_name[27] = "Gambler"
id_to_enemy_name[28] = "Sorcerer"
id_to_enemy_name[29] = "Creeper"
id_to_enemy_name[30] = "Nightwalker"
id_to_enemy_name[32] = "Fortuneteller"
id_to_enemy_name[33] = "Luna Bandit" #silver rock
id_to_enemy_name[34] = "Hot Rod"
id_to_enemy_name[35] = "Cannon Gun" # tornado step
id_to_enemy_name[36] = "Living Bone"
id_to_enemy_name[37] = "Devastator"
id_to_enemy_name[38] = "Lance Soldier"
id_to_enemy_name[39] = "Driller Mole"
id_to_enemy_name[40] = "Shaman"
id_to_enemy_name[41] = "Neoshadow"
id_to_enemy_name[42] = "Magnum Loader"
id_to_enemy_name[43] = "Morning Star"
id_to_enemy_name[44] = "Tornado Step" #Cannon Gun
id_to_enemy_name[45] = "Gargoyle Knight"
id_to_enemy_name[46] = "Gargoyle Warrior"
id_to_enemy_name[47] = "Silver Rock" #Luna Bandit
id_to_enemy_name[48] = "Wight Knight"
id_to_enemy_name[49] = "Emerald Blues"
id_to_enemy_name[50] = "Crimson Jazz"
id_to_enemy_name[51] = "Crescendo"
id_to_enemy_name[52] = "Creeper Plant"
id_to_enemy_name[53] = "Cerberus RC"
id_to_enemy_name[54] = "Thresholder"
id_to_enemy_name[56] = "Possessor"
id_to_enemy_name[60] = "Lock"
id_to_enemy_name[61] = "Shock"
id_to_enemy_name[62] = "Barrel"
id_to_enemy_name[63] = "Air Pirate"
id_to_enemy_name[64] = "Fat Bandit"
id_to_enemy_name[65] = "Fiery Globe"
id_to_enemy_name[66] = "Icy Cube"
id_to_enemy_name[69] = "Aerial Knocker"
id_to_enemy_name[70] = "Small Urn"
id_to_enemy_name[71] = "Big Urn"
# id_to_enemy_name[72] = "" 100% potion
id_to_enemy_name[73] = "Strafer"
# id_to_enemy_name[74] = ""
# id_to_enemy_name[75] = ""
# id_to_enemy_name[76] = ""
# id_to_enemy_name[77] = ""
# id_to_enemy_name[78] = ""
# id_to_enemy_name[79] = ""
# id_to_enemy_name[80] = ""
# id_to_enemy_name[81] = ""
id_to_enemy_name[82] = "Illuminator"
# id_to_enemy_name[83] = ""
# id_to_enemy_name[84] = ""
id_to_enemy_name[85] = "Shadow Stalker Chandelier"
id_to_enemy_name[86] = "Shadow Stalker Pillar"
id_to_enemy_name[87] = "Undead Pirate A"
id_to_enemy_name[88] = "Undead Pirate B"
id_to_enemy_name[89] = "Undead Pirate C"
id_to_enemy_name[90] = "West Wing Armor"
id_to_enemy_name[91] = "LoD Firework"
id_to_enemy_name[92] = "LoD Rocket"
id_to_enemy_name[93] = "LoD Crate"
# id_to_enemy_name[94] = ""
# id_to_enemy_name[95] = ""
id_to_enemy_name[96] = "Bookmaster"
id_to_enemy_name[97] = "Quickplay (Aladdin)"
id_to_enemy_name[98] = "Quickplay (Sora)"
id_to_enemy_name[99] = "Speedster End"
id_to_enemy_name[100] = "Speedster Start"
id_to_enemy_name[102] = "Hyabusa"
id_to_enemy_name[103] = "Abu Ice Crystal"
# id_to_enemy_name[104] = ""
# id_to_enemy_name[105] = ""
# id_to_enemy_name[106] = ""
id_to_enemy_name[107] = "Water Clone"
id_to_enemy_name[108] = "Aladdin Dash"
id_to_enemy_name[109] = "Pan Attack"
id_to_enemy_name[111] = "Stitch Ukulele"
id_to_enemy_name[112] = "Graveyard/Toy Soldier"
# id_to_enemy_name[114] = ""
id_to_enemy_name[115] = "Lance Soldier Idle Hit"
id_to_enemy_name[116] = "Lance Soldier RC Start"
id_to_enemy_name[117] = "Lance Solder Idle Hit (Double)"
id_to_enemy_name[118] = "Lance Soldier RC End"
id_to_enemy_name[119] = "Dusk (Station)"
id_to_enemy_name[120] = "Dusk (STT)"
id_to_enemy_name[121] = "Creeper (STT)"
id_to_enemy_name[122] = "Hayner (Finisher)"
id_to_enemy_name[123] = "Creeper Plant RC"
id_to_enemy_name[125] = "Crescendo RC"
id_to_enemy_name[126] = "Gambler RC"
# id_to_enemy_name[127] = ""
# id_to_enemy_name[128] = ""
id_to_enemy_name[129] = "Meg"
id_to_enemy_name[130] = "Assassin (STT)"
id_to_enemy_name[131] = "Vivi (Finisher)"
id_to_enemy_name[132] = "Setzer (Finisher)"
id_to_enemy_name[133] = "Luxord Minigame"
id_to_enemy_name[134] = "Card"
# id_to_enemy_name[135] = ""
# id_to_enemy_name[136] = ""
# id_to_enemy_name[137] = ""
id_to_enemy_name[138] = "Bulky Vendor (Stage 1/4)"
id_to_enemy_name[139] = "Bulky Vendor (Stage 2/4)"
id_to_enemy_name[140] = "Bulky Vendor (Stage 3/4)"
id_to_enemy_name[141] = "Bulky Vendor (Stage 4/4)"
id_to_enemy_name[142] = "Bulky Vendor Dying"
# id_to_enemy_name[143] = ""
id_to_enemy_name[144] = "Hydra Head"
id_to_enemy_name[145] = "Dusk (STT Day 1)"
# id_to_enemy_name[146] = ""
id_to_enemy_name[147] = "BC Box"
# id_to_enemy_name[148] = ""
id_to_enemy_name[149] = "Junk Breaking"
id_to_enemy_name[150] = "BEES"
# id_to_enemy_name[151] = ""
id_to_enemy_name[152] = "HT Hazards"
id_to_enemy_name[154] = "Junk Hitting Junk"
id_to_enemy_name[155] = "PR Net Hitting"
id_to_enemy_name[156] = "Odd Mushroom 1"
id_to_enemy_name[157] = "Odd Mushroom 2"
id_to_enemy_name[158] = "Odd Mushroom 3"
id_to_enemy_name[159] = "Odd Mushroom 4"
id_to_enemy_name[160] = "Odd Mushroom 5"
id_to_enemy_name[161] = "Even Mushroom 1"
id_to_enemy_name[162] = "Even Mushroom 2"
id_to_enemy_name[163] = "Even Mushroom 3"
id_to_enemy_name[164] = "Even Mushroom 4"
id_to_enemy_name[165] = "Even Mushroom 5"
id_to_enemy_name[166] = "Mushroom Prize 1"
id_to_enemy_name[167] = "Mushroom Prize 2"
id_to_enemy_name[168] = "Mushroom Prize 3"
id_to_enemy_name[169] = "Mushroom Prize 4"
id_to_enemy_name[170] = "Mushroom Prize 5"

id_to_enemy_name[171] = "Befuddler"
id_to_enemy_name[172] = "Camo Cannon"
id_to_enemy_name[173] = "Aerial Viking"
id_to_enemy_name[174] = "Aerial Champ"
id_to_enemy_name[175] = "Necromancer"
id_to_enemy_name[176] = "Magic Phantom"
id_to_enemy_name[177] = "Spring Metal"
id_to_enemy_name[178] = "Runemaster"
id_to_enemy_name[179] = "Iron Hammer"
id_to_enemy_name[180] = "Lance Warrior"
id_to_enemy_name[181] = "Mad Bumper"
id_to_enemy_name[182] = "Reckless"

id_to_enemy_name[183] = "CoR Drive Orb Hit"
id_to_enemy_name[184] = "CoR Drive Orb Final Hit"
id_to_enemy_name[185] = "Valves"
id_to_enemy_name[186] = "Vexen Anti-Sora"
# id_to_enemy_name[187] = ""
id_to_enemy_name[188] = "Zexion Soothe/Herb/Heal/Mend"
id_to_enemy_name[189] = "Zexion Spirit"
id_to_enemy_name[190] = "Zexion Stamina"
id_to_enemy_name[191] = "Zexion Riches/Wealth"
id_to_enemy_name[192] = "Zexion Jackpot/Bounty"
id_to_enemy_name[193] = "Zexion Treasure/Lucky"
id_to_enemy_name[194] = "Zexion Bonus"
id_to_enemy_name[195] = "Seal Magic Break"
id_to_enemy_name[196] = "Seal Attack Break"
id_to_enemy_name[197] = "Seal Magic Break Final Hit"
id_to_enemy_name[198] = "Seal Attack Break Final Hit"


import random
import string
from typing import List

from Class.exceptions import RandomizerExceptions
from Class.seedSettings import ExtraConfigurationData
from Module.RandomizerSettings import RandomizerSettings
from Module.hints import Hints
from Module.multiworld import MultiWorld, MultiWorldConfig
from Module.newRandomize import Randomizer
from Module.seedEvaluation import LocationInformedSeedValidator
from Module.zipper import SeedZip, SeedZipResult


def generateSeed(
    settings: RandomizerSettings, extra_data: ExtraConfigurationData
) -> SeedZipResult:
    last_error = None
    for attempt in range(50):
        newSeedValidation = LocationInformedSeedValidator()
        try:
            randomizer = Randomizer(settings)
            location_spheres = newSeedValidation.validate_seed(
                settings, randomizer
            )
            # hints = Hints.generate_hints(randomizer, settings)
            hints = Hints.generate_hints_v2(randomizer, settings)
            zipper = SeedZip(
                settings, randomizer, hints, extra_data, location_spheres
            )
            return zipper.create_zip()
        except RandomizerExceptions as e:
            characters = string.ascii_letters + string.digits
            settings.random_seed = "".join(random.choice(characters) for i in range(30))
            settings.create_full_seed_string()
            last_error = e
            continue
    raise last_error

def generateSeedCLI(
    settings: RandomizerSettings, extra_data: ExtraConfigurationData
) -> str:
    last_error = None
    for attempt in range(50):
        newSeedValidation = LocationInformedSeedValidator()
        try:
            randomizer = Randomizer(settings)
            location_spheres = newSeedValidation.validate_seed(
                settings, randomizer, False
            )
            # hints = Hints.generate_hints(randomizer, settings)
            hints = Hints.generate_hints_v2(randomizer, settings)
            zipper = SeedZip(
                settings, randomizer, hints, extra_data, location_spheres
            )
            return zipper.make_spoiler_without_zip()
        except RandomizerExceptions as e:
            characters = string.ascii_letters + string.digits
            settings.random_seed = "".join(random.choice(characters) for i in range(30))
            settings.create_full_seed_string()
            last_error = e
            continue
    raise last_error


def generateMultiWorldSeed(
    settingsSet: List[RandomizerSettings], extra_data: ExtraConfigurationData
) -> list[SeedZipResult]:
    newSeedValidation = LocationInformedSeedValidator()
    randomizers = []
    unreachables = []
    last_error = None

    for player_settings in settingsSet:
        for attempt in range(50):
            try:
                last_error = None
                randomizer = Randomizer(player_settings)
                unreachable = newSeedValidation.validate_seed(
                    player_settings, randomizer
                )
                randomizers.append(randomizer)
                unreachables.append(unreachable)

                break
            except RandomizerExceptions as e:
                characters = string.ascii_letters + string.digits
                player_settings.random_seed = "".join(
                    random.choice(characters) for i in range(30)
                )
                player_settings.create_full_seed_string()
                last_error = e
                continue
        if last_error is not None:
            raise last_error

    # each individual randomization is done and valid, now we can mix the item pools
    m = MultiWorld(randomizers, MultiWorldConfig(settingsSet[0]))

    seed_outputs: list[SeedZipResult] = []
    for settings, randomizer, unreachable in zip(
        settingsSet, randomizers, unreachables
    ):
        hints = Hints.generate_hints_v2(randomizer, settings)
        zipper = SeedZip(
            settings, randomizer, hints, extra_data, unreachable, m.multi_output
        )
        seed_outputs.append(zipper.create_zip())

    return seed_outputs





class GeneratorException(Exception):
    pass

class HintException(Exception):
    pass

class SettingsException(Exception):
    pass

class ValidationException(Exception):
    pass

class CantAssignItemException(Exception):
    pass

class BossEnemyException(Exception):
    pass

class BackendException(Exception):
    pass

RandomizerExceptions = (GeneratorException,HintException,SettingsException,ValidationException)


from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Consumable(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.ITEM)


Potion = Consumable(1, "Potion")
HiPotion = Consumable(2, "Hi-Potion")
Ether = Consumable(3, "Ether")
Elixir = Consumable(4, "Elixir")
MegaPotion = Consumable(5, "Mega-Potion")
MegaEther = Consumable(6, "Mega-Ether")
Megalixir = Consumable(7, "Megalixir")
Tent = Consumable(131, "Tent")
DriveRecovery = Consumable(274, "Drive Recovery")
HighDriveRecovery = Consumable(275, "High Drive Recovery")
PowerBoost = Consumable(276, "Power Boost")
MagicBoost = Consumable(277, "Magic Boost")
DefenseBoost = Consumable(278, "Defense Boost")
ApBoost = Consumable(279, "AP Boost")




import random
from typing import List

from Class.itemClass import KH2Item
from Class.newLocationClass import KH2Location
from List.ItemList import Items
from List.configDict import itemType, locationCategory
from Module.RandomizerSettings import RandomizerSettings
from Module.newRandomize import Randomizer, ItemAssignment


class MultiWorldConfig():
    def __init__(self,settings: RandomizerSettings):
        self.mixed_item_count = 20
        self.num_players = 2
        self.item_swaps = [itemType.FIRE, itemType.BLIZZARD, itemType.THUNDER, itemType.CURE, itemType.MAGNET, itemType.REFLECT, itemType.PROMISE_CHARM, itemType.PROOF_OF_NONEXISTENCE, itemType.PROOF_OF_PEACE, itemType.PROOF_OF_CONNECTION, itemType.STORYUNLOCK]
        self.location_swaps = [locationCategory.CHEST]

class MultiWorldEntry():
    def __init__(self, finder, owner, where_found: KH2Location, what_found: KH2Item):
        self.finder = finder
        self.owner = owner
        self.where_found : KH2Location = where_found
        self.what_found : KH2Item = what_found

    def __call__(self):
        location_category = ""
        if self.where_found.LocationCategory in [locationCategory.ITEMBONUS,locationCategory.HYBRIDBONUS,locationCategory.STATBONUS,locationCategory.DOUBLEBONUS]:
            location_category = "BONUS"
        else:
            location_category = self.where_found.LocationCategory.name
        return {"name":self.what_found.Id,"location":location_category+"-"+str(self.where_found.LocationId),"to":self.owner,"from":self.finder}

class MultiWorldOutput():
    def __init__(self):
        self.new_assignments = []

    def add_item(self,item,location,from_player,to_player):
        self.new_assignments.append(MultiWorldEntry(from_player,to_player,location,item))


    def __call__(self):
        output = {"items":[]}
        for a in self.new_assignments:
            output["items"].append(a())
        return output

class MultiWorld():
    def __init__(self, seeds: List[Randomizer], config: MultiWorldConfig):
        self.all_candidate_swaps: List[List[ItemAssignment]] = []

        for seed in seeds:
            self.all_candidate_swaps.append([])
            for assignment in seed.assignments:
                if assignment.location.LocationCategory in config.location_swaps:
                    if assignment.item is not None and assignment.item.ItemType in config.item_swaps:
                        self.all_candidate_swaps[-1].append(assignment)
                    # elif assignment.item2 is not None and assignment.item2.ItemType in config.item_swaps:
                    #     self.all_candidate_swaps[-1].append(assignment)

        for player in range(len(self.all_candidate_swaps)):
            candiate_swaps = self.all_candidate_swaps[player]
            print(f"Player {player}")
            for assignment in candiate_swaps:
                print(f"--- ({assignment.item},{assignment.item2}) in {assignment.location.Description}({assignment.location.LocationId})")

        self.mix(config)

        print("------------------")

        for player in range(len(self.all_candidate_swaps)):
            candiate_swaps = self.all_candidate_swaps[player]
            print(f"Player {player}")
            for assignment in candiate_swaps:
                print(f"--- ({assignment.item},{assignment.item2}) in {assignment.location.Description}({assignment.location.LocationId})")

    

    def mix(self,config: MultiWorldConfig):
        full_swaps = []
        max_item_swap = 999999
        for i,player in enumerate(self.all_candidate_swaps):
            max_item_swap = min(len(player),max_item_swap)
            for j in range(len(player)):
                full_swaps.append((i,j))

        max_item_swap=min(config.mixed_item_count,max_item_swap*config.num_players)
        

        swap_chain = []
        swap_chain.append(random.choice(full_swaps))
        full_swaps.remove(swap_chain[-1])

        for _ in range(max_item_swap-1):
            filtered_list = [it for it in full_swaps if it[0]!=swap_chain[-1][0]]
            swap_chain.append(random.choice(filtered_list))
            full_swaps.remove(swap_chain[-1])

        self.multi_output = MultiWorldOutput()

        prev_owner = swap_chain[-1][0]
        temp_item = swap_chain[-1][1]

        for ch in swap_chain:
            item = self.all_candidate_swaps[prev_owner][temp_item].item
            location = self.all_candidate_swaps[ch[0]][ch[1]].location

            self.multi_output.add_item(item,location,ch[0],prev_owner)
            prev_owner = ch[0]
            temp_item = ch[1]
    
        print(self.multi_output())


        for s in self.multi_output.new_assignments:
            # print(s())
            for ass in self.all_candidate_swaps[s.finder]:
                # print(f"{ass.location == s.where_found} {ass.location}")
                if ass.location == s.where_found:
                    # TODO(zaktherobot)
                    # ass.item = Items.sharedMultiItem()
                    break





from enum import Enum

from List.configDict import locationType
from List.inventory import misc, magic, keyblade, ability, summon
from List.location.graph import RequirementEdge, chest, popup, item_bonus, stat_bonus, \
    LocationGraphBuilder, START_NODE, DefaultLogicGraph
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    AgrabahMapPopup = "Agrabah Map Popup"
    Agrabah = "Agrabah"
    AgrabahChests = "Agrabah Chests"
    Bazaar = "Bazaar"
    BazaarChests = "Bazaar Chests"
    PalaceWalls = "Palace Walls"
    PalaceWallsChests = "Palace Walls Chests"
    CaveOfWondersEntrance = "Cave of Wonders Entrance"
    CaveOfWondersEntranceChests = "Cave of Wonders Entrance Chests"
    ValleyOfStone = "Valley of Stone"
    ValleyOfStoneChests = "Valley of Stone Chests"
    AbuEscort = "Abu Escort"
    ChasmOfChallenges = "Chasm of Challenges"
    ChasmOfChallengesChests = "Chasm of Challenges Chests"
    TreasureRoomBonus = "Treasure Room Bonus"
    TreasureRoom = "Treasure Room"
    TreasureRoomChests = "Treasure Room Chests"
    ElementalLords = "Elemental Lords"
    RuinedChamber = "Ruined Chamber"
    RuinedChamberChests = "Ruined Chamber Chests"
    GenieJafar = "Genie Jafar"
    Lexaeus = "AS Lexaeus"
    DataLexaeus = "Data Lexaeus"


class CheckLocation(str, Enum):
    AgrabahMap = "Agrabah Map"
    AgrabahDarkShard = "Agrabah Dark Shard"
    AgrabahMythrilShard = "Agrabah Mythril Shard"
    AgrabahHiPotion = "Agrabah Hi-Potion"
    AgrabahApBoost = "Agrabah AP Boost"
    AgrabahMythrilStone = "Agrabah Mythril Stone"
    AgrabahMythrilShard2 = "Agrabah Mythril Shard (2)"
    AgrabahSerenityShard = "Agrabah Serenity Shard"
    BazaarMythrilGem = "Bazaar Mythril Gem"
    BazaarPowerShard = "Bazaar Power Shard"
    BazaarHiPotion = "Bazaar Hi-Potion"
    BazaarApBoost = "Bazaar AP Boost"
    BazaarMythrilShard = "Bazaar Mythril Shard"
    PalaceWallsSkillRing = "Palace Walls Skill Ring"
    PalaceWallsMythrilStone = "Palace Walls Mythril Stone"
    CaveEntrancePowerStone = "Cave Entrance Power Stone"
    CaveEntranceMythrilShard = "Cave Entrance Mythril Shard"
    ValleyOfStoneMythrilStone = "Valley of Stone Mythril Stone"
    ValleyOfStoneApBoost = "Valley of Stone AP Boost"
    ValleyOfStoneMythrilShard = "Valley of Stone Mythril Shard"
    ValleyOfStoneHiPotion = "Valley of Stone Hi-Potion"
    AbuEscort = "Abu Escort"
    ChasmOfChallengesCaveOfWondersMap = "Chasm of Challenges Cave of Wonders Map"
    ChasmOfChallengesApBoost = "Chasm of Challenges AP Boost"
    TreasureRoomBonus = "Treasure Room"
    TreasureRoomApBoost = "Treasure Room AP Boost"
    TreasureRoomSerenityGem = "Treasure Room Serenity Gem"
    ElementalLordsBonus = "Elemental Lords"
    LampCharm = "Lamp Charm"
    RuinedChamberTornPages = "Ruined Chamber Torn Pages"
    RuinedChamberRuinsMap = "Ruined Chamber Ruins Map"
    GenieJafarBonus = "Genie Jafar"
    WishingLamp = "Wishing Lamp"
    LexaeusBonus = "Lexaeus Bonus"
    LexaeusStrengthBeyondStrength = "Lexaeus (AS) Strength Beyond Strength"
    DataLexaeus = "Lexaeus (Data) Lost Illusion"

class AGLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_ag_keyblade(inv)
        self.logic[NodeId.Agrabah][NodeId.AgrabahChests] = keyblade_lambda
        self.logic[NodeId.Bazaar][NodeId.BazaarChests] = keyblade_lambda
        self.logic[NodeId.PalaceWalls][NodeId.PalaceWallsChests] = keyblade_lambda
        self.logic[NodeId.CaveOfWondersEntrance][NodeId.CaveOfWondersEntranceChests] = keyblade_lambda
        self.logic[NodeId.ValleyOfStone][NodeId.ValleyOfStoneChests] = keyblade_lambda
        self.logic[NodeId.ChasmOfChallenges][NodeId.ChasmOfChallengesChests] = keyblade_lambda
        self.logic[NodeId.TreasureRoom][NodeId.TreasureRoomChests] = keyblade_lambda
        self.logic[NodeId.RuinedChamber][NodeId.RuinedChamberChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.AgrabahMapPopup] = ItemPlacementHelpers.ag1_check
            self.logic[NodeId.ElementalLords][NodeId.RuinedChamber] = lambda inv : ItemPlacementHelpers.ag2_check(inv) and ItemPlacementHelpers.need_fire_blizzard_thunder(inv)
        else:
            self.logic[START_NODE][NodeId.Agrabah] = ItemPlacementHelpers.ag1_check
            self.logic[NodeId.PalaceWalls][NodeId.RuinedChamber] = ItemPlacementHelpers.need_fire_blizzard_thunder
            self.logic[NodeId.Lexaeus][NodeId.AgrabahMapPopup] = ItemPlacementHelpers.ag2_check


def make_graph(graph: LocationGraphBuilder):
    ag = locationType.Agrabah
    ag_logic = AGLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(ag_logic)

    agrabah_map_popup = graph.add_location(NodeId.AgrabahMapPopup, [
        popup(353, CheckLocation.AgrabahMap, ag),
    ])
    agrabah_chests = graph.add_location(NodeId.AgrabahChests, [
        chest(28, CheckLocation.AgrabahDarkShard, ag),
        chest(29, CheckLocation.AgrabahMythrilShard, ag),
        chest(30, CheckLocation.AgrabahHiPotion, ag),
        chest(132, CheckLocation.AgrabahApBoost, ag),
        chest(133, CheckLocation.AgrabahMythrilStone, ag),
        chest(249, CheckLocation.AgrabahMythrilShard2, ag),
        chest(501, CheckLocation.AgrabahSerenityShard, ag),
    ])
    agrabah = graph.add_location(NodeId.Agrabah, [])
    bazaar_chests = graph.add_location(NodeId.BazaarChests, [
        chest(31, CheckLocation.BazaarMythrilGem, ag),
        chest(32, CheckLocation.BazaarPowerShard, ag),
        chest(33, CheckLocation.BazaarHiPotion, ag),
        chest(134, CheckLocation.BazaarApBoost, ag),
        chest(135, CheckLocation.BazaarMythrilShard, ag),
    ])
    bazaar = graph.add_location(NodeId.Bazaar, [])
    palace_walls_chests = graph.add_location(NodeId.PalaceWallsChests, [
        chest(136, CheckLocation.PalaceWallsSkillRing, ag),
        chest(520, CheckLocation.PalaceWallsMythrilStone, ag),
    ])
    palace_walls = graph.add_location(NodeId.PalaceWalls, [])
    cave_of_wonders_entrance_chests = graph.add_location(NodeId.CaveOfWondersEntranceChests, [
        chest(250, CheckLocation.CaveEntrancePowerStone, ag),
        chest(251, CheckLocation.CaveEntranceMythrilShard, ag),
    ])
    cave_of_wonders_entrance = graph.add_location(NodeId.CaveOfWondersEntrance, [])
    valley_of_stone_chests = graph.add_location(NodeId.ValleyOfStoneChests, [
        chest(35, CheckLocation.ValleyOfStoneMythrilStone, ag),
        chest(36, CheckLocation.ValleyOfStoneApBoost, ag),
        chest(137, CheckLocation.ValleyOfStoneMythrilShard, ag),
        chest(138, CheckLocation.ValleyOfStoneHiPotion, ag),
    ])
    valley_of_stone = graph.add_location(NodeId.ValleyOfStone, [])
    abu_escort = graph.add_location(NodeId.AbuEscort, [
        item_bonus(42, CheckLocation.AbuEscort, ag, vanilla=ability.SummonBoost),
    ])
    chasm_of_challenges_chests = graph.add_location(NodeId.ChasmOfChallengesChests, [
        chest(487, CheckLocation.ChasmOfChallengesCaveOfWondersMap, ag),
        chest(37, CheckLocation.ChasmOfChallengesApBoost, ag),
    ])
    chasm_of_challenges = graph.add_location(NodeId.ChasmOfChallenges, [])
    treasure_room_bonus = graph.add_location(NodeId.TreasureRoomBonus, [
        stat_bonus(46, CheckLocation.TreasureRoomBonus, ag),
    ])
    treasure_room_chests = graph.add_location(NodeId.TreasureRoomChests, [
        chest(502, CheckLocation.TreasureRoomApBoost, ag),
        chest(503, CheckLocation.TreasureRoomSerenityGem, ag),
    ])
    treasure_room = graph.add_location(NodeId.TreasureRoom, [])
    elemental_lords = graph.add_location(NodeId.ElementalLords, [
        item_bonus(37, CheckLocation.ElementalLordsBonus, ag, vanilla=ability.FinishingLeap),
        popup(300, CheckLocation.LampCharm, ag, vanilla=summon.LampCharm),
    ])
    ruined_chamber_chests = graph.add_location(NodeId.RuinedChamberChests, [
        chest(34, CheckLocation.RuinedChamberTornPages, ag, vanilla=misc.TornPages),
        chest(486, CheckLocation.RuinedChamberRuinsMap, ag),
    ])
    ruined_chamber = graph.add_location(NodeId.RuinedChamber, [])
    genie_jafar = graph.add_location(NodeId.GenieJafar, [
        item_bonus(15, CheckLocation.GenieJafarBonus, ag, vanilla=magic.Fire),
        popup(303, CheckLocation.WishingLamp, ag, vanilla=keyblade.WishingLamp),
    ])
    lexaeus = graph.add_location(NodeId.Lexaeus, [
        stat_bonus(65, CheckLocation.LexaeusBonus, [ag, locationType.AS]),
        popup(545, CheckLocation.LexaeusStrengthBeyondStrength, [ag, locationType.AS]),
    ])
    data_lexaeus = graph.add_location(NodeId.DataLexaeus, [
        popup(550, CheckLocation.DataLexaeus, [ag, locationType.DataOrg]),
    ])

    graph.register_superboss(data_lexaeus)

    graph.add_edge(agrabah, agrabah_chests)
    graph.add_edge(bazaar, bazaar_chests)
    graph.add_edge(palace_walls, palace_walls_chests)
    graph.add_edge(cave_of_wonders_entrance, cave_of_wonders_entrance_chests)
    graph.add_edge(valley_of_stone, valley_of_stone_chests)
    graph.add_edge(chasm_of_challenges, chasm_of_challenges_chests)
    graph.add_edge(treasure_room, treasure_room_chests)
    graph.add_edge(ruined_chamber, ruined_chamber_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, agrabah_map_popup)
        graph.add_edge(agrabah_map_popup, agrabah)
        graph.add_edge(agrabah, bazaar)
        graph.add_edge(bazaar, palace_walls)
        graph.add_edge(palace_walls, cave_of_wonders_entrance)
        graph.add_edge(cave_of_wonders_entrance, valley_of_stone)
        graph.add_edge(valley_of_stone, abu_escort)
        graph.add_edge(abu_escort, chasm_of_challenges, RequirementEdge(battle=True))
        graph.add_edge(chasm_of_challenges, treasure_room_bonus, RequirementEdge(battle=True))
        graph.add_edge(treasure_room_bonus, treasure_room)
        graph.add_edge(treasure_room, elemental_lords, RequirementEdge(battle=True))
        graph.add_edge(elemental_lords, ruined_chamber, RequirementEdge(battle=True))
        graph.add_edge(ruined_chamber, genie_jafar, RequirementEdge(battle=True))
        graph.add_edge(genie_jafar, lexaeus, RequirementEdge(battle=True))
        graph.add_edge(lexaeus, data_lexaeus)
        graph.register_superboss(lexaeus)
        graph.register_first_boss(elemental_lords)
        graph.register_last_story_boss(genie_jafar)
    else:
        graph.add_edge(START_NODE, agrabah)
        graph.add_edge(agrabah, bazaar)
        graph.add_edge(bazaar, palace_walls)
        graph.add_edge(palace_walls, ruined_chamber,RequirementEdge(battle=True))
        graph.add_edge(ruined_chamber, genie_jafar, RequirementEdge(battle=True))
        graph.add_edge(genie_jafar, lexaeus, RequirementEdge(battle=True))
        graph.add_edge(lexaeus, agrabah_map_popup)
        graph.add_edge(agrabah_map_popup, cave_of_wonders_entrance)
        graph.add_edge(cave_of_wonders_entrance, valley_of_stone)
        graph.add_edge(valley_of_stone, abu_escort)
        graph.add_edge(abu_escort, chasm_of_challenges, RequirementEdge(battle=True))
        graph.add_edge(chasm_of_challenges, treasure_room_bonus, RequirementEdge(battle=True))
        graph.add_edge(treasure_room_bonus, treasure_room)
        graph.add_edge(treasure_room, elemental_lords, RequirementEdge(battle=True))
        graph.add_edge(elemental_lords, data_lexaeus, RequirementEdge(battle=True))
        graph.register_first_boss(genie_jafar)
        graph.register_last_story_boss(elemental_lords)



from dataclasses import dataclass

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class MiscItem(InventoryItem):
    id: int
    name: str
    type: itemType


NullItem = MiscItem(0, "", itemType.SYNTH)
TornPages = MiscItem(32, "Torn Pages", itemType.TORN_PAGE)
MunnyPouchOlette = MiscItem(362, "Munny Pouch", itemType.MUNNY_POUCH)
ObjectiveItem = MiscItem(363, "Objective Completion Item", itemType.OBJECTIVE) # previously crystal orb
EmblemItem = MiscItem(363, "Emblem", itemType.OBJECTIVE) # previously crystal orb
SeifersTrophy = MiscItem(364, "Seifer's Trophy", itemType.KEYITEM)
Poster = MiscItem(366, "Poster", itemType.KEYITEM)
OlympusStone = MiscItem(370, "Olympus Stone", itemType.OCSTONE)
AuronsStatue = MiscItem(371, "Auron's Statue", itemType.KEYITEM)
CursedMedallion = MiscItem(372, "Cursed Medallion", itemType.KEYITEM)
Present = MiscItem(373, "Presents", itemType.KEYITEM)
DecoyPresents = MiscItem(374, "Decoy Presents", itemType.KEYITEM)
ObjectiveReport = MiscItem(239, "Objective Report", itemType.KEYITEM)
UnknownDisk = MiscItem(462, "Unknown Disk", itemType.MANUFACTORYUNLOCK)  # Dummy 15
PromiseCharm = MiscItem(524, "PromiseCharm", itemType.PROMISE_CHARM)
MunnyPouchMickey = MiscItem(535, "Munny Pouch", itemType.MUNNY_POUCH)
HadesCupTrophy = MiscItem(537, "Hades Cup Trophy", itemType.TROPHY)
StruggleTrophy = MiscItem(538, '"The Struggle" Trophy', itemType.KEYITEM)



from enum import Enum

from Class.newLocationClass import KH2Location
from List.configDict import locationType, locationCategory, itemType
from List.location.graph import DefaultLogicGraph, RequirementEdge, LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    AwakeningPuzzle = "Awakening Puzzle"
    HeartPuzzle = "Heart Puzzle"
    DualityPuzzle = "Duality Puzzle"
    FrontierPuzzle = "Frontier Puzzle"
    DaylightPuzzle = "Daylight Puzzle"
    SunsetPuzzle = "Sunset Puzzle"


class CheckLocation(str, Enum):
    AwakeningApBoost = "Awakening (AP Boost)"
    HeartSerenityCrystal = "Heart (Serenity Crystal)"
    DualityRareDocument = "Duality (Rare Document)"
    FrontierManifestIllusion = "Frontier (Manifest Illusion)"
    DaylightExecutivesRing = "Daylight (Executive's Ring)"
    SunsetGrandRibbon = "Sunset (Grand Ribbon)"

class PuzzleLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblades_unlock_chests):
        DefaultLogicGraph.__init__(self,NodeId)

        if not reverse_rando:
            def awakening_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.hb1_check(inv)
                    and ItemPlacementHelpers.bc1_check(inv)
                    and ItemPlacementHelpers.tt1_check(inv)
                    and ItemPlacementHelpers.lod1_check(inv)
                )

            def heart_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.dc2_check(inv)
                    and ItemPlacementHelpers.oc1_check(inv)
                    and ItemPlacementHelpers.pr1_check(inv)
                    and ItemPlacementHelpers.hb1_check(inv)
                    and ItemPlacementHelpers.lod1_check(inv)
                )

            def duality_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.ht1_check(inv)
                    and ItemPlacementHelpers.dc2_check(inv)
                    and ItemPlacementHelpers.pr1_check(inv)
                    and ItemPlacementHelpers.ag1_check(inv)
                    and ItemPlacementHelpers.oc1_check(inv)
                    and ItemPlacementHelpers.hb1_check(inv)
                )

            def frontier_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.tt2_check(inv)
                    and ItemPlacementHelpers.ht1_check(inv)
                    and ItemPlacementHelpers.hb2_check(inv)
                    and ItemPlacementHelpers.dc2_check(inv)
                    and ItemPlacementHelpers.ag1_check(inv)
                    and ItemPlacementHelpers.pr1_check(inv)
                    and ItemPlacementHelpers.lod1_check(inv)
                )
            def daylight_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.need_torn_pages(5)(inv)
                    and ItemPlacementHelpers.hb2_check(inv)
                    and ItemPlacementHelpers.lod2_check(inv)
                    and ItemPlacementHelpers.tt3_check(inv)
                    and ItemPlacementHelpers.pr2_check(inv)
                    and ItemPlacementHelpers.ag2_check(inv)
                    and ItemPlacementHelpers.pl1_check(inv)
                    and ItemPlacementHelpers.ht1_check(inv)
                    and ItemPlacementHelpers.oc1_check(inv)
                    and ItemPlacementHelpers.sp1_check(inv)
                    and ItemPlacementHelpers.bc1_check(inv)
                    and ItemPlacementHelpers.twtnw_roxas_check(inv)
                )

            def sunset_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.hb2_check(inv)
                    and ItemPlacementHelpers.tt3_check(inv)
                    and ItemPlacementHelpers.sp2_check(inv)
                    and ItemPlacementHelpers.pr2_check(inv)
                    and ItemPlacementHelpers.ag2_check(inv)
                    and ItemPlacementHelpers.bc2_check(inv)
                    and ItemPlacementHelpers.pl1_check(inv)
                    and ItemPlacementHelpers.lod1_check(inv)
                    and ItemPlacementHelpers.twtnw_post_saix_check(inv)
                )
            self.logic[START_NODE][NodeId.AwakeningPuzzle] = awakening_checker
            self.logic[START_NODE][NodeId.HeartPuzzle] = heart_checker
            self.logic[START_NODE][NodeId.DualityPuzzle] = duality_checker
            self.logic[START_NODE][NodeId.FrontierPuzzle] = frontier_checker
            self.logic[START_NODE][NodeId.DaylightPuzzle] = daylight_checker
            self.logic[START_NODE][NodeId.SunsetPuzzle] = sunset_checker
        else:
            def awakening_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.bc2_check(inv)
                    and ItemPlacementHelpers.tt3_check(inv)
                )

            def heart_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.oc2_check(inv)
                    and ItemPlacementHelpers.pr2_check(inv)
                )

            def duality_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.pr2_check(inv)
                    and ItemPlacementHelpers.oc2_check(inv)
                )

            def frontier_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.pr2_check(inv)
                    and ItemPlacementHelpers.ag2_check(inv)
                    and ItemPlacementHelpers.need_fire_blizzard_thunder(inv)
                )

            def daylight_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.need_torn_pages(5)(inv)
                    and ItemPlacementHelpers.pl2_check(inv)
                    and ItemPlacementHelpers.ag2_check(inv)
                    and ItemPlacementHelpers.need_fire_blizzard_thunder(inv)
                    and ItemPlacementHelpers.hb2_check(inv)
                )

            def sunset_checker(inv: list[int]) -> bool:
                return (
                    ItemPlacementHelpers.need_growths(inv)
                    and ItemPlacementHelpers.hb2_check(inv)
                    and ItemPlacementHelpers.tt3_check(inv)
                    and ItemPlacementHelpers.ag2_check(inv)
                    and ItemPlacementHelpers.need_fire_blizzard_thunder(inv)
                )
            self.logic[START_NODE][NodeId.AwakeningPuzzle] = awakening_checker
            self.logic[START_NODE][NodeId.HeartPuzzle] = heart_checker
            self.logic[START_NODE][NodeId.DualityPuzzle] = duality_checker
            self.logic[START_NODE][NodeId.FrontierPuzzle] = frontier_checker
            self.logic[START_NODE][NodeId.DaylightPuzzle] = daylight_checker
            self.logic[START_NODE][NodeId.SunsetPuzzle] = sunset_checker

def puzzle_reward(
    loc_id: int, description: str, invalid_checks: list[itemType]
) -> KH2Location:
    return KH2Location(
        LocationId=loc_id,
        Description=description,
        LocationCategory=locationCategory.CREATION,
        LocationTypes=[locationType.Puzzle, locationType.Creations],
        InvalidChecks=invalid_checks,
    )


def make_graph(graph: LocationGraphBuilder):
    puzzle_logic = PuzzleLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(puzzle_logic)
    awakening = graph.add_location(
        NodeId.AwakeningPuzzle,
        [
            puzzle_reward(
                0, CheckLocation.AwakeningApBoost, invalid_checks=[itemType.REPORT]
            ),
        ],
    )
    heart = graph.add_location(
        NodeId.HeartPuzzle,
        [
            puzzle_reward(
                1, CheckLocation.HeartSerenityCrystal, invalid_checks=[itemType.REPORT]
            ),
        ],
    )
    duality = graph.add_location(
        NodeId.DualityPuzzle,
        [
            puzzle_reward(
                2, CheckLocation.DualityRareDocument, invalid_checks=[itemType.REPORT]
            ),
        ],
    )
    frontier = graph.add_location(
        NodeId.FrontierPuzzle,
        [
            puzzle_reward(
                3,
                CheckLocation.FrontierManifestIllusion,
                invalid_checks=[itemType.REPORT],
            ),
        ],
    )
    daylight = graph.add_location(
        NodeId.DaylightPuzzle,
        [
            puzzle_reward(
                4,
                CheckLocation.DaylightExecutivesRing,
                invalid_checks=[itemType.TORN_PAGE, itemType.REPORT],
            ),
        ],
    )
    sunset = graph.add_location(
        NodeId.SunsetPuzzle,
        [
            puzzle_reward(
                5, CheckLocation.SunsetGrandRibbon, invalid_checks=[itemType.REPORT]
            ),
        ],
    )
    graph.add_edge(START_NODE, awakening)
    graph.add_edge(START_NODE, heart)
    graph.add_edge(START_NODE, duality)
    graph.add_edge(START_NODE, frontier)
    graph.add_edge(START_NODE, daylight)
    graph.add_edge(START_NODE, sunset)


from enum import Enum

from List.configDict import locationType
from List.inventory import misc, magic, keyblade, ability
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, item_bonus, hybrid_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    BambooGrove = "Bamboo Grove"
    BambooGroveChests = "Bamboo Grove Chests"
    EncampmentAreaMap = "Encampment Area Map"
    Mission3 = "Mission 3"
    Checkpoint = "Checkpoint"
    CheckpointChests = "Checkpoint Chests"
    MountainTrail = "Mountain Trail"
    MountainTrailChests = "Mountain Trail Chests"
    VillageCaveMapPopup = "Village Cave Map Popup"
    VillageCave = "Village Cave"
    VillageCaveChests = "Village Cave Chests"
    VillageCaveBonus = "Village Cave Bonus"
    Ridge = "Ridge"
    RidgeChests = "Ridge Chests"
    ShanYu = "Shan-Yu"
    ThroneRoom = "Throne Room"
    ThroneRoomChests = "Throne Room Chests"
    StormRider = "Storm Rider"
    DataXigbar = "Data Xigbar"


class CheckLocation(str, Enum):
    BambooGroveDarkShard = "Bamboo Grove Dark Shard"
    BambooGroveEther = "Bamboo Grove Ether"
    BambooGroveMythrilShard = "Bamboo Grove Mythril Shard"
    EncampmentAreaMap = "Encampment Area Map"
    Mission3 = "Mission 3"
    CheckpointHiPotion = "Checkpoint Hi-Potion"
    CheckpointMythrilShard = "Checkpoint Mythril Shard"
    MountainTrailLightningShard = "Mountain Trail Lightning Shard"
    MountainTrailRecoveryRecipe = "Mountain Trail Recovery Recipe"
    MountainTrailEther = "Mountain Trail Ether"
    MountainTrailMythrilShard = "Mountain Trail Mythril Shard"
    VillageCaveAreaMap = "Village Cave Area Map"
    VillageCaveApBoost = "Village Cave AP Boost"
    VillageCaveDarkShard = "Village Cave Dark Shard"
    VillageCaveBonus = "Village Cave Bonus"
    RidgeFrostShard = "Ridge Frost Shard"
    RidgeApBoost = "Ridge AP Boost"
    ShanYuBonus = "Shan-Yu"
    HiddenDragon = "Hidden Dragon"
    ThroneRoomTornPages = "Throne Room Torn Pages"
    ThroneRoomPalaceMap = "Throne Room Palace Map"
    ThroneRoomApBoost = "Throne Room AP Boost"
    ThroneRoomQueenRecipe = "Throne Room Queen Recipe"
    ThroneRoomApBoost2 = "Throne Room AP Boost (2)"
    ThroneRoomOgreShield = "Throne Room Ogre Shield"
    ThroneRoomMythrilCrystal = "Throne Room Mythril Crystal"
    ThroneRoomOrichalcum = "Throne Room Orichalcum"
    StormRiderBonus = "Storm Rider"
    DataXigbarDefenseBoost = "Xigbar (Data) Defense Boost"

class LoDLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_lod_keyblade(inv)
        self.logic[NodeId.BambooGrove][NodeId.BambooGroveChests] = keyblade_lambda
        self.logic[NodeId.Checkpoint][NodeId.CheckpointChests] = keyblade_lambda
        self.logic[NodeId.MountainTrail][NodeId.MountainTrailChests] = keyblade_lambda
        self.logic[NodeId.VillageCave][NodeId.VillageCaveChests] = keyblade_lambda
        self.logic[NodeId.Ridge][NodeId.RidgeChests] = keyblade_lambda
        self.logic[NodeId.ThroneRoom][NodeId.ThroneRoomChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.BambooGrove] = ItemPlacementHelpers.lod1_check
            self.logic[NodeId.ShanYu][NodeId.ThroneRoom] = ItemPlacementHelpers.lod2_check
        else:
            self.logic[START_NODE][NodeId.MountainTrail] = ItemPlacementHelpers.lod1_check
            self.logic[NodeId.StormRider][NodeId.BambooGrove] = ItemPlacementHelpers.lod2_check

def make_graph(graph: LocationGraphBuilder):
    lod = locationType.LoD
    lod_logic = LoDLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(lod_logic)

    bamboo_grove_chests = graph.add_location(NodeId.BambooGroveChests, [
        chest(245, CheckLocation.BambooGroveDarkShard, lod),
        chest(497, CheckLocation.BambooGroveEther, lod),
        chest(498, CheckLocation.BambooGroveMythrilShard, lod),
    ])
    bamboo_grove = graph.add_location(NodeId.BambooGrove, [])
    encampment_area_map = graph.add_location(NodeId.EncampmentAreaMap, [
        popup(350, CheckLocation.EncampmentAreaMap, lod),
    ])
    mission3 = graph.add_location(NodeId.Mission3, [
        popup(417, CheckLocation.Mission3, lod),
    ])
    checkpoint_chests = graph.add_location(NodeId.CheckpointChests, [
        chest(21, CheckLocation.CheckpointHiPotion, lod),
        chest(121, CheckLocation.CheckpointMythrilShard, lod),
    ])
    checkpoint = graph.add_location(NodeId.Checkpoint, [])
    mountain_trail_chests = graph.add_location(NodeId.MountainTrailChests, [
        chest(22, CheckLocation.MountainTrailLightningShard, lod),
        chest(23, CheckLocation.MountainTrailRecoveryRecipe, lod),
        chest(122, CheckLocation.MountainTrailEther, lod),
        chest(123, CheckLocation.MountainTrailMythrilShard, lod),
    ])
    mountain_trail = graph.add_location(NodeId.MountainTrail, [])
    village_cave_map_popup = graph.add_location(NodeId.VillageCaveMapPopup, [
        popup(495, CheckLocation.VillageCaveAreaMap, lod),
    ])
    village_cave_chests = graph.add_location(NodeId.VillageCaveChests, [
        chest(124, CheckLocation.VillageCaveApBoost, lod),
        chest(125, CheckLocation.VillageCaveDarkShard, lod),
    ])
    village_cave = graph.add_location(NodeId.VillageCave, [])
    village_cave_bonus = graph.add_location(NodeId.VillageCaveBonus, [
        item_bonus(43, CheckLocation.VillageCaveBonus, lod, vanilla=ability.SlideDash),
    ])
    ridge_chests = graph.add_location(NodeId.RidgeChests, [
        chest(24, CheckLocation.RidgeFrostShard, lod),
        chest(126, CheckLocation.RidgeApBoost, lod),
    ])
    ridge = graph.add_location(NodeId.Ridge, [])
    shan_yu = graph.add_location(NodeId.ShanYu, [
        hybrid_bonus(9, CheckLocation.ShanYuBonus, lod, vanilla=ability.AerialSweep),
        popup(257, CheckLocation.HiddenDragon, lod, vanilla=keyblade.HiddenDragon),
    ])
    throne_room_chests = graph.add_location(NodeId.ThroneRoomChests, [
        chest(25, CheckLocation.ThroneRoomTornPages, lod, vanilla=misc.TornPages),
        chest(127, CheckLocation.ThroneRoomPalaceMap, lod),
        chest(26, CheckLocation.ThroneRoomApBoost, lod),
        chest(27, CheckLocation.ThroneRoomQueenRecipe, lod),
        chest(128, CheckLocation.ThroneRoomApBoost2, lod),
        chest(129, CheckLocation.ThroneRoomOgreShield, lod),
        chest(130, CheckLocation.ThroneRoomMythrilCrystal, lod),
        chest(131, CheckLocation.ThroneRoomOrichalcum, lod),
    ])
    throne_room = graph.add_location(NodeId.ThroneRoom, [])
    storm_rider = graph.add_location(NodeId.StormRider, [
        item_bonus(10, CheckLocation.StormRiderBonus, lod, vanilla=magic.Thunder),
    ])
    data_xigbar = graph.add_location(NodeId.DataXigbar, [
        popup(555, CheckLocation.DataXigbarDefenseBoost, [lod, locationType.DataOrg]),
    ])

    graph.register_superboss(data_xigbar)
    
    graph.add_edge(bamboo_grove, bamboo_grove_chests)
    graph.add_edge(checkpoint, checkpoint_chests)
    graph.add_edge(mountain_trail, mountain_trail_chests)
    graph.add_edge(village_cave, village_cave_chests)
    graph.add_edge(ridge, ridge_chests)
    graph.add_edge(throne_room, throne_room_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, bamboo_grove)
        graph.add_edge(bamboo_grove, encampment_area_map)
        graph.add_edge(encampment_area_map, mission3, RequirementEdge(battle=True))
        graph.add_edge(encampment_area_map, checkpoint)
        graph.add_edge(mission3, mountain_trail, RequirementEdge(battle=True))
        graph.add_edge(mountain_trail, village_cave_map_popup, RequirementEdge(battle=True))
        graph.add_edge(village_cave_map_popup, village_cave)
        graph.add_edge(village_cave, village_cave_bonus, RequirementEdge(battle=True))
        graph.add_edge(village_cave_bonus, ridge)
        graph.add_edge(ridge, shan_yu, RequirementEdge(battle=True))
        graph.add_edge(shan_yu, throne_room, RequirementEdge(battle=True, req=ItemPlacementHelpers.lod2_check))
        graph.add_edge(throne_room, storm_rider, RequirementEdge(battle=True))
        graph.add_edge(storm_rider, data_xigbar, RequirementEdge(battle=True))
        graph.register_first_boss(shan_yu)
        graph.register_last_story_boss(storm_rider)
    else:
        graph.add_edge(START_NODE, mountain_trail)
        graph.add_edge(mountain_trail, checkpoint)
        graph.add_edge(mountain_trail, ridge)
        graph.add_edge(ridge, throne_room, RequirementEdge(battle=True))
        graph.add_edge(throne_room, storm_rider, RequirementEdge(battle=True))
        graph.add_edge(storm_rider, bamboo_grove, RequirementEdge(req=ItemPlacementHelpers.lod2_check))
        graph.add_edge(bamboo_grove, encampment_area_map, RequirementEdge(battle=True))
        graph.add_edge(encampment_area_map, mission3)
        graph.add_edge(mission3, village_cave_map_popup, RequirementEdge(battle=True))
        graph.add_edge(village_cave_map_popup, village_cave)
        graph.add_edge(village_cave, village_cave_bonus, RequirementEdge(battle=True))
        graph.add_edge(village_cave_bonus, shan_yu, RequirementEdge(battle=True))
        graph.add_edge(shan_yu, data_xigbar, RequirementEdge(battle=True))
        graph.register_first_boss(storm_rider)
        graph.register_last_story_boss(shan_yu)



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Staff(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.STAFF)
    stat_entry: int


MeteorStaff = Staff(150, "Meteor Staff", stat_entry=89)
RisingDragon = Staff(154, "Rising Dragon", stat_entry=93)
NobodyLance = Staff(155, "Nobody Lance", stat_entry=94)
ShamansRelic = Staff(156, "Shaman's Relic", stat_entry=95)
SaveTheQueenPlus = Staff(503, "Save The Queen+", stat_entry=146)
PreciousMushroom = Staff(549, "Precious Mushroom", stat_entry=154)
PreciousMushroomPlus = Staff(550, "Precious Mushroom+", stat_entry=155)
PremiumMushroom = Staff(551, "Premium Mushroom", stat_entry=156)
CenturionPlus = Staff(546, "Centurion+", stat_entry=151)



from typing import Optional

from Class.newLocationClass import KH2Location
from List.LvupStats import DreamWeaponOffsets
from List.configDict import locationCategory, locationType
from List.inventory import ability
from List.inventory.item import InventoryItem
from List.location.graph import LocationGraphBuilder, START_NODE, RequirementEdge
from Module.itemPlacementRestriction import ItemPlacementHelpers


def level_reward(level: int, description: str, vanilla: Optional[InventoryItem]) -> KH2Location:
    vanilla_items: list[int] = []
    if vanilla is not None:
        vanilla_items.append(vanilla.id)
    return KH2Location(level, description, locationCategory.LEVEL, [locationType.Level], VanillaItems=vanilla_items)


def make_graph(graph: LocationGraphBuilder):
    settings = graph.settings
    max_level = settings.max_level_checks

    vanilla_levels: dict[int, dict[int, InventoryItem]] = {
        50: {
            2: ability.ComboBoost,
            4: ability.ExperienceBoost,
            7: ability.MagicLockOn,
            9: ability.ReactionBoost,
            10: ability.ItemBoost,
            12: ability.LeafBracer,
            14: ability.FireBoost,
            15: ability.DriveBoost,
            17: ability.Draw,
            20: ability.CombinationBoost,
            23: ability.DamageDrive,
            25: ability.AirComboBoost,
            28: ability.BlizzardBoost,
            30: ability.DriveConverter,
            32: ability.NegativeCombo,
            34: ability.OnceMore,
            36: ability.FinishingPlus,
            39: ability.ThunderBoost,
            41: ability.Defender,
            44: ability.BerserkCharge,
            46: ability.Jackpot,
            48: ability.SecondChance,
            50: ability.DamageControl
        },
        99: {
            7: ability.ComboBoost,
            9: ability.ExperienceBoost,
            12: ability.MagicLockOn,
            15: ability.ReactionBoost,
            17: ability.ItemBoost,
            20: ability.LeafBracer,
            23: ability.FireBoost,
            25: ability.DriveBoost,
            28: ability.Draw,
            31: ability.CombinationBoost,
            33: ability.DamageDrive,
            36: ability.AirComboBoost,
            39: ability.BlizzardBoost,
            41: ability.DriveConverter,
            44: ability.NegativeCombo,
            47: ability.OnceMore,
            49: ability.FinishingPlus,
            53: ability.ThunderBoost,
            59: ability.Defender,
            65: ability.BerserkCharge,
            73: ability.Jackpot,
            85: ability.SecondChance,
            99: ability.DamageControl
        }
    }

    node_index = 0
    locations: list[str] = []
    current_location_list: list[KH2Location] = []

    already_added_level_to_group = False
    level_offsets = DreamWeaponOffsets()
    excluded_levels = settings.excluded_levels()
    for level in range(1, 100):
        level_description = f"Level {level}"

        if settings.split_levels:
            shield = level_offsets.get_shield_level(max_level, level)
            staff = level_offsets.get_staff_level(max_level, level)
            if shield and staff:
                level_description = f"Level Sw: {level} Sh: {shield} St: {staff}"

        vanilla_item: Optional[InventoryItem] = None
        if max_level in vanilla_levels and level in vanilla_levels[max_level]:
            vanilla_item = vanilla_levels[max_level][level]
        elif level in vanilla_levels[99]:
            vanilla_item = vanilla_levels[99][level]

        current_location_list.append(level_reward(level, level_description, vanilla_item))

        if level not in excluded_levels:
            if already_added_level_to_group:
                locations.append(graph.add_location(f"LevelGroup-{node_index}", current_location_list))
                current_location_list = []
                node_index += 1
                already_added_level_to_group = False
            else:
                already_added_level_to_group = True

    # Add in the last group if there are any stragglers
    if len(current_location_list) > 0:
        locations.append(graph.add_location(f"LevelGroup-{node_index}", current_location_list))

    for index, location in enumerate(locations):
        if index == 0:
            graph.add_edge(START_NODE, location, RequirementEdge(req=ItemPlacementHelpers.make_level_group_check(index)))
        else:
            graph.add_edge(locations[index - 1], location, RequirementEdge(req=ItemPlacementHelpers.make_level_group_check(index)))



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class AnsemReport(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.REPORT)
    report_number: int


AnsemReport1 = AnsemReport(226, "Secret Ansem's Report 1", report_number=1)
AnsemReport2 = AnsemReport(227, "Secret Ansem's Report 2", report_number=2)
AnsemReport3 = AnsemReport(228, "Secret Ansem's Report 3", report_number=3)
AnsemReport4 = AnsemReport(229, "Secret Ansem's Report 4", report_number=4)
AnsemReport5 = AnsemReport(230, "Secret Ansem's Report 5", report_number=5)
AnsemReport6 = AnsemReport(231, "Secret Ansem's Report 6", report_number=6)
AnsemReport7 = AnsemReport(232, "Secret Ansem's Report 7", report_number=7)
AnsemReport8 = AnsemReport(233, "Secret Ansem's Report 8", report_number=8)
AnsemReport9 = AnsemReport(234, "Secret Ansem's Report 9", report_number=9)
AnsemReport10 = AnsemReport(235, "Secret Ansem's Report 10", report_number=10)
AnsemReport11 = AnsemReport(236, "Secret Ansem's Report 11", report_number=11)
AnsemReport12 = AnsemReport(237, "Secret Ansem's Report 12", report_number=12)
AnsemReport13 = AnsemReport(238, "Secret Ansem's Report 13", report_number=13)


def all_reports() -> list[AnsemReport]:
    return [AnsemReport1, AnsemReport2, AnsemReport3, AnsemReport4, AnsemReport5, AnsemReport6, AnsemReport7,
            AnsemReport8, AnsemReport9, AnsemReport10, AnsemReport11, AnsemReport12, AnsemReport13]



from enum import Enum

from List.configDict import locationType
from List.inventory import keyblade, magic
from List.location.graph import RequirementEdge, popup, LocationGraphBuilder, START_NODE, DefaultLogicGraph
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    AtlanticaTutorial = "Atlantica Tutorial"
    Ursula = "Ursula"
    NewDayIsDawning = "New Day is Dawning"


class CheckLocation(str, Enum):
    UnderseaKingdomMap = "Undersea Kingdom Map"
    MysteriousAbyss = "Mysterious Abyss"
    MusicalBlizzardElement = "Musical Blizzard Element"
    MusicalOrichalcumPlus = "Musical Orichalcum+"


class ATLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando):
        DefaultLogicGraph.__init__(self,NodeId)
        if not reverse_rando:
            self.logic[NodeId.AtlanticaTutorial][NodeId.Ursula] = ItemPlacementHelpers.need_2_magnet
            self.logic[NodeId.Ursula][NodeId.NewDayIsDawning] = ItemPlacementHelpers.need_3_thunders
        else:
            self.logic[NodeId.NewDayIsDawning][NodeId.Ursula] = ItemPlacementHelpers.need_1_magnet
            self.logic[NodeId.Ursula][NodeId.AtlanticaTutorial] = lambda inv: ItemPlacementHelpers.need_2_magnet(inv) and ItemPlacementHelpers.need_3_thunders(inv)


def make_graph(graph: LocationGraphBuilder):
    at = locationType.Atlantica
    at_logic = ATLogicGraph(graph.reverse_rando)
    graph.add_logic(at_logic)

    tutorial = graph.add_location(NodeId.AtlanticaTutorial, [
        popup(367, CheckLocation.UnderseaKingdomMap, at),
    ])
    ursula = graph.add_location(NodeId.Ursula, [
        popup(287, CheckLocation.MysteriousAbyss, at, vanilla=keyblade.MysteriousAbyss),
    ])
    new_day_is_dawning = graph.add_location(NodeId.NewDayIsDawning, [
        popup(279, CheckLocation.MusicalBlizzardElement, at, vanilla=magic.Blizzard),
        popup(538, CheckLocation.MusicalOrichalcumPlus, at),
    ])

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, tutorial)
        graph.add_edge(tutorial, ursula, RequirementEdge())
        graph.add_edge(ursula, new_day_is_dawning, RequirementEdge())
    else:
        graph.add_edge(START_NODE, new_day_is_dawning)
        graph.add_edge(new_day_is_dawning, ursula, RequirementEdge())
        graph.add_edge(ursula, tutorial, RequirementEdge())



from random import random
from bisect import bisect_left
from itertools import accumulate


def weighted_sample_without_replacement(population, weights, k):
    wts = list(weights)
    sampl = []
    rnums = [random() for _ in range(k)]
    for r in rnums:
        acm_wts = list(accumulate(wts))
        total = acm_wts[-1]
        i = bisect_left(acm_wts, total * r)
        p = population[i]
        wts[i] = 0
        sampl.append(p)
    return sampl



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem
from List.location import weaponslot


@dataclass(frozen=True)
class Keyblade(InventoryItem):
    id: int
    name: str
    weaponslot_id: int
    type: itemType = field(init=False, default=itemType.KEYBLADE)


Oathkeeper = Keyblade(42, "Oathkeeper", weaponslot_id=weaponslot.LocationId.Oathkeeper)
Oblivion = Keyblade(43, "Oblivion", weaponslot_id=weaponslot.LocationId.Oblivion)
Pureblood = Keyblade(71, "Pureblood", weaponslot_id=weaponslot.LocationId.Pureblood)  # Anti-Form Dummy
StarSeeker = Keyblade(480, "Star Seeker", weaponslot_id=weaponslot.LocationId.StarSeeker)
HiddenDragon = Keyblade(481, "Hidden Dragon", weaponslot_id=weaponslot.LocationId.HiddenDragon)
HerosCrest = Keyblade(484, "Hero's Crest", weaponslot_id=weaponslot.LocationId.HerosCrest)
Monochrome = Keyblade(485, "Monochrome", weaponslot_id=weaponslot.LocationId.Monochrome)
FollowTheWind = Keyblade(486, "Follow the Wind", weaponslot_id=weaponslot.LocationId.FollowTheWind)
CircleOfLife = Keyblade(487, "Circle of Life", weaponslot_id=weaponslot.LocationId.CircleOfLife)
PhotonDebugger = Keyblade(488, "Photon Debugger", weaponslot_id=weaponslot.LocationId.PhotonDebugger)
GullWing = Keyblade(489, "Gull Wing", weaponslot_id=weaponslot.LocationId.GullWing)
RumblingRose = Keyblade(490, "Rumbling Rose", weaponslot_id=weaponslot.LocationId.RumblingRose)
GuardianSoul = Keyblade(491, "Guardian Soul", weaponslot_id=weaponslot.LocationId.GuardianSoul)
WishingLamp = Keyblade(492, "Wishing Lamp", weaponslot_id=weaponslot.LocationId.WishingLamp)
DecisivePumpkin = Keyblade(493, "Decisive Pumpkin", weaponslot_id=weaponslot.LocationId.DecisivePumpkin)
SleepingLion = Keyblade(494, "Sleeping Lion", weaponslot_id=weaponslot.LocationId.SleepingLion)
SweetMemories = Keyblade(495, "Sweet Memories", weaponslot_id=weaponslot.LocationId.SweetMemories)
MysteriousAbyss = Keyblade(496, "Mysterious Abyss", weaponslot_id=weaponslot.LocationId.MysteriousAbyss)
FatalCrest = Keyblade(497, "Fatal Crest", weaponslot_id=weaponslot.LocationId.FatalCrest)
BondOfFlame = Keyblade(498, "Bond of Flame", weaponslot_id=weaponslot.LocationId.BondOfFlame)
Fenrir = Keyblade(499, "Fenrir", weaponslot_id=weaponslot.LocationId.Fenrir)
UltimaWeapon = Keyblade(500, "Ultima Weapon", weaponslot_id=weaponslot.LocationId.UltimaWeapon)
TwoBecomeOne = Keyblade(543, "Two Become One", weaponslot_id=weaponslot.LocationId.TwoBecomeOne)
WinnersProof = Keyblade(544, "Winner's Proof", weaponslot_id=weaponslot.LocationId.WinnersProof)

def get_all_keyblades() -> list[Keyblade]:
    return get_locking_keyblades() + [
        Oblivion, 
        Pureblood,
        StarSeeker,
        GullWing,
        GuardianSoul,
        MysteriousAbyss,
        FatalCrest,
        Fenrir,
        UltimaWeapon,
    ]


def get_locking_keyblade_names() -> list[str]:
    return [i.name for i in get_locking_keyblades()]


def get_locking_keyblades() -> list[Keyblade]:
    return [Oathkeeper,
            HiddenDragon,
            HerosCrest,
            Monochrome,
            FollowTheWind,
            CircleOfLife,
            PhotonDebugger,
            WishingLamp,
            DecisivePumpkin,
            SleepingLion,
            SweetMemories,
            BondOfFlame,
            TwoBecomeOne,
            WinnersProof,
            RumblingRose,
            ]


import os
import random
from pathlib import Path

from Class.openkhmod import Asset
from List import configDict
from Module import appconfig


class EndingPictureRandomizer:

    @staticmethod
    def endingpic_rando_options() -> dict[str, str]:
        return {
            configDict.VANILLA: "Vanilla",
            configDict.RANDOMIZE_IN_GAME_ONLY: "Randomize (in-game only)",
            configDict.RANDOMIZE_CUSTOM_ONLY: "Randomize (custom only)",
            configDict.RANDOMIZE_ALL: "Randomize (in-game + custom)",
        }

    @staticmethod
    def randomize_end_screen(setting: str) -> list[Asset]:
        """Randomizes the ending screen, returning a list of assets to be added to a mod."""

        assets: list[Asset] = []
        if setting == configDict.VANILLA:
            return assets

        vanilla_by_locale: dict[str, list[str]] = {
            "fr": [
                "remastered/menu/fr/ending.2ld/FR_ending_2ld2.png",
                "remastered/menu/fr/ending.2ld/FR_ending_2ld3.png",
                "remastered/menu/fr/ending.2ld/FR_ending_2ld4.png",
                "remastered/menu/fr/ending.2ld/FR_ending_2ld5.png",
                "remastered/menu/fr/ending.2ld/FR_ending_2ld6.png",
                "remastered/menu/fr/ending.2ld/FR_ending_2ld7.png",
            ],
            "gr": [
                "remastered/menu/gr/ending.2ld/GR_ending_2ld2.png",
                "remastered/menu/gr/ending.2ld/GR_ending_2ld3.png",
                "remastered/menu/gr/ending.2ld/GR_ending_2ld4.png",
                "remastered/menu/gr/ending.2ld/GR_ending_2ld5.png",
                "remastered/menu/gr/ending.2ld/GR_ending_2ld6.png",
                "remastered/menu/gr/ending.2ld/GR_ending_2ld7.png",
            ],
            "it": [
                "remastered/menu/it/ending.2ld/IT_ending_2ld2.png",
                "remastered/menu/it/ending.2ld/IT_ending_2ld3.png",
                "remastered/menu/it/ending.2ld/IT_ending_2ld4.png",
                "remastered/menu/it/ending.2ld/IT_ending_2ld5.png",
                "remastered/menu/it/ending.2ld/IT_ending_2ld6.png",
                "remastered/menu/it/ending.2ld/IT_ending_2ld7.png",
            ],
            "sp": [
                "remastered/menu/sp/ending.2ld/SP_ending_2ld2.png",
                "remastered/menu/sp/ending.2ld/SP_ending_2ld3.png",
                "remastered/menu/sp/ending.2ld/SP_ending_2ld4.png",
                "remastered/menu/sp/ending.2ld/SP_ending_2ld5.png",
                "remastered/menu/sp/ending.2ld/SP_ending_2ld6.png",
                "remastered/menu/sp/ending.2ld/SP_ending_2ld7.png",
            ],
            "us": [
                "remastered/menu/us/ending.2ld/US_ending_2ld2.png",
                "remastered/menu/us/ending.2ld/US_ending_2ld3.png",
                "remastered/menu/us/ending.2ld/US_ending_2ld4.png",
                "remastered/menu/us/ending.2ld/US_ending_2ld5.png",
                "remastered/menu/us/ending.2ld/US_ending_2ld6.png",
                "remastered/menu/us/ending.2ld/US_ending_2ld7.png",
            ],
        }

        custom = EndingPictureRandomizer._collect_custom_endpics()

        if setting == configDict.RANDOMIZE_ALL:
            if len(custom) == 0:
                # No custom ones, so fall back to in-game
                setting = configDict.RANDOMIZE_IN_GAME_ONLY
            else:
                # We're only choosing one in the end, so flip a coin to choose between in-game one or custom
                coinflip = random.randint(0, 1)
                if coinflip == 0:
                    setting = configDict.RANDOMIZE_IN_GAME_ONLY
                else:
                    setting = configDict.RANDOMIZE_CUSTOM_ONLY

        if setting == configDict.RANDOMIZE_IN_GAME_ONLY:
            for endpic_list in vanilla_by_locale.values():
                choice = random.choice(endpic_list)
                assets.append({
                    "name": endpic_list[0],
                    "platform": "pc",
                    "multi": [{"name": endpic} for endpic in endpic_list[1:]],
                    "method": "copy",
                    "source": [{
                        "name": choice,
                        "type": "internal",
                    }]
                })

        if setting == configDict.RANDOMIZE_CUSTOM_ONLY and len(custom) > 0:
            choice = random.choice(custom)
            for endpic_list in vanilla_by_locale.values():
                assets.append({
                    "name": endpic_list[0],
                    "platform": "pc",
                    "multi": [{"name": endpic} for endpic in endpic_list[1:]],
                    "method": "copy",
                    "source": [{"name": str(choice)}]
                })

        return assets

    @staticmethod
    def _collect_custom_endpics() -> list[Path]:
        result: list[Path] = []

        custom_visuals_path = appconfig.read_custom_visuals_path()
        if custom_visuals_path is not None:
            ending_pictures_path = custom_visuals_path / "ending-pictures"
            if ending_pictures_path.is_dir():
                for root, dirs, files in os.walk(ending_pictures_path):
                    root_path = Path(root)
                    for file in files:
                        name, extension = os.path.splitext(file)
                        if extension.lower() == ".png":
                            file_path = root_path / file
                            result.append(file_path)

        return result
